from dataclasses import dataclass
import numpy as np
import pandas as pd
import os
import re
import math
import warnings


class EventResults:
    """
    Class to store results (summary) of the RTModel fitting for a single event. It reads the Nature.txt file and also extracts
    the best fit names of each class of models (PS, PX, BS, BO, LS, LX, LO) and their chi2 values.
    """

    def __init__(self, event_folder_):
        """ Initialize the EventResults class with the event folder path."""
        self.event_folder = event_folder_
        self.event_name = event_folder_.split('/')[-1]

        # Initialize other attributes that need function below
        self.ps_best = BestModel()
        self.px_best = BestModel()
        self.bs_best = BestModel()
        self.bo_best = BestModel()
        self.ls_best = BestModel()
        self.lx_best = BestModel()
        self.lo_best = BestModel()
        self.complete_classification = None
        self.final_models = None
        self.number_of_final_models = None
        self.found_a_planet = None
        self.found_a_2L1S_solution = None

    def extract_information_from_nature_file(self):
        """Extracts information from the Nature.txt file in the event folder."""
        nature_file = os.path.join(self.event_folder, "Nature.txt")
        with open(nature_file, "r") as f:
            lines = f.readlines()
            content = ''.join(lines)
        # Extract the line starting with 'Successful:'
        self.complete_classification = ""
        match = re.search(r"Successful:\s*(.*)", content)
        if match:
            self.complete_classification = match.group(1).strip()
        self.found_a_planet = "Planetary lens" in self.complete_classification

        # Find the start of the 'chisquare model' section
        try:
            start = next(i for i, line in enumerate(lines) if "chisquare" in line)
        except StopIteration:
            start = None
        # Extract the best fit names
        self.final_models = []
        if start is not None:
            for line in lines[start + 1:]:
                parts = line.split()
                if len(parts) != 2 or not parts[1].endswith(".txt"):
                    break
                self.final_models.append(parts[1].replace(".txt", ""))
        self.number_of_final_models = len(self.final_models)

        # Check if any model starts with 'L'
        self.found_a_2L1S_solution = any(m.startswith("L") for m in self.final_models)

        # Extract chi2 values
        # Initialize a dictionary to hold chi2 values
        self.best_chi2 = {}
        # Pattern for the models you want
        chi2_pattern = re.compile(r"^(PS|PX|BS|BO|LS|LX|LO)\s*[:=]\s*([\d.eE+-]+)")
        for line in lines:
            match = chi2_pattern.search(line)
            if match:
                model, value = match.groups()
                self.best_chi2[model] = float(value)


    def looking_for_the_names_of_best_model_of_each_category(self):
        """Looks for the names of the best model of each category in the alternative models."""
        files = [f for f in os.listdir(self.event_folder + '/Models') if f.endswith(".txt")]

        for file_name in files:
            # Load model and get chi2
            model_name = file_name.split('.')[0]
            category = model_name[0:2]  # Get the first two characters to determine the category
            model_path = os.path.join(self.event_folder + '/Models', file_name)
            model = ModelResults(model_path)
            chi2 = model.model_parameters.chi2

            best_object = getattr(self, f"{category.lower()}_best")
            best_object.chi2 = self.best_chi2.get(category)

            if math.isclose(chi2, self.best_chi2[category], rel_tol=1e-5):
                best_object.chi2 = chi2
                best_object.name = model_name
            elif chi2 < self.best_chi2[category]:
                warnings.warn(
                    f"Chi2 for {file_name} ({chi2}) differs from existing {category} self.best_chi2[category] ({self.best_chi2[category]})\n"
                    f"It probably means that the chi2 value in the Nature.txt file is not the same as the one in the Model folder."
                )
                # If the chi2 is better than the current best, update the best object
                best_object.chi2 = chi2
                best_object.name = model_name


@dataclass
class BestModel:
    def __init__(self, name=None, chi2=None):
        self.name = name
        self.chi2 = chi2


class ModelResults:
    def __init__(self, file_to_be_read, *, data_challenge_lc_number=None):
        self.model_type = file_to_be_read.split('/')[-1][0:2]
        self.data_challenge_lc_number = data_challenge_lc_number
        if self.model_type == 'PS':
            self.model_type_extensive_name = 'Single Lens Single Source (1L1S)'
            self.model_parameters = SingleLensSingleSourcePS(file_to_be_read)
        elif self.model_type == 'PX':
            self.model_type_extensive_name = 'Single Lens Single Source with Parallax (1L1S+plx)'
            self.model_parameters = SingleLensSingleSourceWithParallaxPX(file_to_be_read)
        elif self.model_type == 'BS':
            self.model_type_extensive_name = 'Single Lens Binary Source (1L2S)'
            self.model_parameters = SingleLensBinarySourceBS(file_to_be_read)
        elif self.model_type == 'BO':
            self.model_type_extensive_name = 'Single Lens Binary Source with Xallarap (1L2S+xlp)'
            self.model_parameters = SingleLensBinarySourceWithXallarapBO(file_to_be_read)
        elif self.model_type == 'LS':
            self.model_type_extensive_name = 'Binary Lens Single Source (2L1S)'
            self.model_parameters = BinaryLensSingleSourceLS(file_to_be_read)
        elif self.model_type == 'LX':
            self.model_type_extensive_name = 'Binary Lens Single Source with Parallax (2L1S+plx)'
            self.model_parameters = BinaryLensSingleSourceWithParallaxLX(file_to_be_read)
        elif self.model_type == 'LO':
            self.model_type_extensive_name = 'Binary Lens Single Source with Parallax and Orbital Motion (2L1S+plx+OM)'
            self.model_parameters = BinaryLensSingleSourceWithOrbitalMotionLO(file_to_be_read)
        else:
            raise ValueError(f"Model type {self.model_type} not recognized.")
        self.covariance_matrix = pd.read_csv(file_to_be_read, skiprows=2, header=None, sep=r'\s+')


@dataclass
class SingleLensSingleSourcePS:
    """
    PS	Single_lens_single_source
    4 parameters
    """
    number_of_parameters: int
    u0: float  # Impact parameter normalized to Einstein angle (internally fit in logarithmic (ln) scale)
    u0_error: float
    tE: float  # Einstein time in days (internally fit in logarithmic (ln) scale)
    tE_error: float
    t0: float  # Closest approach time in HJD
    t0_error: float
    rho: float  # Source radius normalized to Einstein angle (internally fit in logarithmic (ln) scale)
    rho_error: float
    t_star: float  # Source crossing time in days
    t_star_error: float
    chi2: float  # Chi2 value for the fit
    blends: float
    sources: float
    blendings: float
    blendings_error: float
    baselines: float
    baselines_error: float

    def __init__(self, file_to_be_read):
        self.number_of_parameters = 4
        with open(file_to_be_read, 'r') as f:
            lines = f.readlines()
            parameters = lines[0].split(' ')
            errors = lines[1].split(' ')
            self.u0 = float(parameters[0])
            self.u0_error = float(errors[0])
            self.tE = float(parameters[1])
            self.tE_error = float(errors[1])
            self.t0 = float(parameters[2])
            self.t0_error = float(errors[2])
            self.rho = float(parameters[3])
            self.rho_error = float(errors[3])
            self.t_star = self.rho * self.tE
            self.t_star_error = np.sqrt(self.tE**2 * self.rho_error**2 + self.rho**2 * self.tE_error**2)
            self.chi2 = float(parameters[-1])

            self.blends = np.array(
                [float(parameters[i]) for i in range(self.number_of_parameters, len(parameters) - 1, 2)])
            blends_error = [float(errors[i]) for i in range(self.number_of_parameters, len(errors), 2)]
            self.sources = np.array(
                [float(parameters[i]) for i in range(self.number_of_parameters + 1, len(parameters) - 1, 2)])
            sources_error = [float(errors[i]) for i in range(self.number_of_parameters + 1, len(errors), 2)]

            # TODO CHECK THIS
            self.blendings = self.blends / (self.sources + 1.e-12 * self.blends)
            self.blendings_error = blends_error
            self.baselines = -2.5 * np.log10(self.blends + self.sources)
            self.baselines_error = sources_error


@dataclass
class SingleLensSingleSourceWithParallaxPX:
    """
    PX	Single_lens_single_source with parallax
    6 parameters
    """
    number_of_parameters: int
    u0: float  # Impact parameter normalized to Einstein angle
    u0_error: float
    tE: float  # Einstein time in days (internally fit in logarithmic (ln) scale)
    tE_error: float
    t0: float  # Closest approach time in HJD
    t0_error: float
    rho: float  # Source radius normalized to Einstein angle (internally fit in logarithmic (ln) scale)
    rho_error: float
    t_star: float  # Source crossing time in days
    t_star_error: float
    piEN: float  # Parallax component along North
    piEN_error: float
    piEE: float  # Parallax component along East
    piEE_error: float
    chi2: float  # Chi2 value for the fit
    blends: float
    sources: float
    blendings: float
    blendings_error: float
    baselines: float
    baselines_error: float

    def __init__(self, file_to_be_read):
        self.number_of_parameters = 6
        with open(file_to_be_read, 'r') as f:
            lines = f.readlines()
            parameters = lines[0].split(' ')
            errors = lines[1].split(' ')
            self.u0 = float(parameters[0])
            self.u0_error = float(errors[0])
            self.tE = float(parameters[1])
            self.tE_error = float(errors[1])
            self.t0 = float(parameters[2])
            self.t0_error = float(errors[2])
            self.rho = float(parameters[3])
            self.rho_error = float(errors[3])
            self.t_star = self.rho * self.tE
            self.t_star_error = np.sqrt(self.tE**2 * self.rho_error**2 + self.rho**2 * self.tE_error**2)
            self.piEN = float(parameters[4])
            self.piEN_error = float(errors[4])
            self.piEE = float(parameters[5])
            self.piEE_error = float(errors[5])
            self.chi2 = float(parameters[-1])

            self.blends = np.array(
                [float(parameters[i]) for i in range(self.number_of_parameters, len(parameters) - 1, 2)])
            blends_error = [float(errors[i]) for i in range(self.number_of_parameters, len(errors), 2)]
            self.sources = np.array(
                [float(parameters[i]) for i in range(self.number_of_parameters + 1, len(parameters) - 1, 2)])
            sources_error = [float(errors[i]) for i in range(self.number_of_parameters + 1, len(errors), 2)]

            # TODO CHECK THIS
            self.blendings = self.blends / (self.sources + 1.e-12 * self.blends)
            self.blendings_error = blends_error
            self.baselines = -2.5 * np.log10(self.blends + self.sources)
            self.baselines_error = sources_error


@dataclass
class SingleLensBinarySourceBS:
    """
    BS	Single_lens_binary_source
    7 parameters
    """
    number_of_parameters: int
    tE: float  # Einstein time in days (internally fit in logarithmic (ln) scale)
    tE_error: float
    flux_ratio: float  # Flux ratio of the secondary to the primary source (internally fit in logarithmic (ln) scale)
    flux_ratio_error: float
    u01: float  # Impact parameter normalized to Einstein angle for primary source
    u01_error: float
    u02: float  # Impact parameter of the secondary source
    u02_error: float
    t01: float  # Closest approach time of the primary source
    t01_error: float
    t02: float  # Closest approach time of the secondary source
    t02_error: float
    rho1: float  # Source radius for the primary source (internally fit in logarithmic (ln) scale)
    rho1_error: float
    t_star1: float  # Source 1 crossing time in days
    t_star1_error: float
    chi2: float  # Chi2 value for the fit
    blends: float
    sources: float
    blendings: float
    blendings_error: float
    baselines: float
    baselines_error: float

    def __init__(self, file_to_be_read):
        self.number_of_parameters = 7
        with open(file_to_be_read, 'r') as f:
            lines = f.readlines()
            parameters = lines[0].split(' ')
            errors = lines[1].split(' ')
            self.tE = float(parameters[0])
            self.tE_error = float(errors[0])
            self.flux_ratio = float(parameters[1])
            self.flux_ratio_error = float(errors[1])
            self.u01 = float(parameters[2])
            self.u01_error = float(errors[2])
            self.u02 = float(parameters[3])
            self.u02_error = float(errors[3])
            self.t01 = float(parameters[4])
            self.t01_error = float(errors[4])
            self.t02 = float(parameters[5])
            self.t02_error = float(errors[5])
            self.rho1 = float(parameters[6])
            self.rho1_error = float(errors[6])
            self.t_star1 = self.rho1 * self.tE
            self.t_star1_error = np.sqrt(self.tE**2 * self.rho1_error**2 + self.rho1**2 * self.tE_error**2)
            self.chi2 = float(parameters[-1])

            self.blends = np.array(
                [float(parameters[i]) for i in range(self.number_of_parameters, len(parameters) - 1, 2)])
            blends_error = [float(errors[i]) for i in range(self.number_of_parameters, len(errors), 2)]
            self.sources = np.array(
                [float(parameters[i]) for i in range(self.number_of_parameters + 1, len(parameters) - 1, 2)])
            sources_error = [float(errors[i]) for i in range(self.number_of_parameters + 1, len(errors), 2)]

            # TODO CHECK THIS
            self.blendings = self.blends / (self.sources + 1.e-12 * self.blends)
            self.blendings_error = blends_error
            self.baselines = -2.5 * np.log10(self.blends + self.sources)
            self.baselines_error = sources_error


@dataclass
class SingleLensBinarySourceWithXallarapBO:
    """
    BO	Single_lens_binary_source with xallarap
    10 parameters
    """
    number_of_parameters: int
    u01: float  # Impact parameter of the primary source
    t01: float  # Closest approach time of the primary source
    tE: float  # Einstein time in days (internally fit in logarithmic (ln) scale)
    rho1: float  # Source radius for the primary source (internally fit in logarithmic (ln) scale)
    t_star1: float  # Source 1 crossing time in days
    t_star1_error: float
    xi1: float  # Xallarap component parallel to the source velocity
    xi2: float  # Xallarap component orthogonal to the source velocity
    omega: float  # Orbital angular velocity in days^-1
    inc: float  # Inclination of the orbital plane in radians
    phi: float  # Phase of the orbit from the passage on the line of nodes
    qs: float  # Mass ratio of the secondary to the primary source
    chi2: float  # Chi2 value for the fit
    blends: float
    sources: float
    blendings: float
    blendings_error: float
    baselines: float
    baselines_error: float

    def __init__(self, file_to_be_read):
        self.number_of_parameters = 10
        with open(file_to_be_read, 'r') as f:
            lines = f.readlines()
            parameters = lines[0].split(' ')
            errors = lines[1].split(' ')
            self.u01 = float(parameters[0])
            self.u01_error = float(errors[0])
            self.t01 = float(parameters[1])
            self.t01_error = float(errors[1])
            self.tE = float(parameters[2])
            self.tE_error = float(errors[2])
            self.rho1 = float(parameters[3])
            self.rho1_error = float(errors[3])
            self.t_star1 = self.rho1 * self.tE
            self.t_star1_error = np.sqrt(self.tE**2 * self.rho1_error**2 + self.rho1**2 * self.tE_error**2)
            self.xi1 = float(parameters[4])
            self.xi1_error = float(errors[4])
            self.xi2 = float(parameters[5])
            self.xi2_error = float(errors[5])
            self.omega = float(parameters[6])
            self.omega_error = float(errors[6])
            self.inc = float(parameters[7])
            self.inc_error = float(errors[7])
            self.phi = float(parameters[8])
            self.phi_error = float(errors[8])
            self.qs = float(parameters[9])
            self.qs_error = float(errors[9])
            self.chi2 = float(parameters[-1])

            self.blends = np.array(
                [float(parameters[i]) for i in range(self.number_of_parameters, len(parameters) - 1, 2)])
            blends_error = [float(errors[i]) for i in range(self.number_of_parameters, len(errors), 2)]
            self.sources = np.array(
                [float(parameters[i]) for i in range(self.number_of_parameters + 1, len(parameters) - 1, 2)])
            sources_error = [float(errors[i]) for i in range(self.number_of_parameters + 1, len(errors), 2)]

            # TODO CHECK THIS
            self.blendings = self.blends / (self.sources + 1.e-12 * self.blends)
            self.blendings_error = blends_error
            self.baselines = -2.5 * np.log10(self.blends + self.sources)
            self.baselines_error = sources_error


@dataclass
class BinaryLensSingleSourceLS:
    """
    LS	Binary_lens_single_source
    7 parameters
    """
    number_of_parameters: int
    separation: float  # Separation between the lenses in Einstein radii (internally fit in logarithmic (ln) scale)
    separation_error: float
    mass_ratio: float  # Mass ratio of the secondary to the primary lens (internally fit in logarithmic (ln) scale)
    mass_ratio_error: float
    u0: float  # Impact parameter normalized to Einstein angle
    u0_error: float
    alpha: float  # Angle between the source velocity and the vector pointing from the secondary to the primary lens
    alpha_error: float
    rho: float  # Source radius normalized to Einstein angle (internally fit in logarithmic (ln) scale)
    rho_error: float
    tE: float  # Einstein time in days (internally fit in logarithmic (ln) scale)
    tE_error: float
    t_star: float  # Source crossing time in days
    t_star_error: float
    t0: float  # Closest approach time in HJD to the barycenter
    t0_error: float
    chi2: float  # Chi2 value for the fit
    blends: float
    sources: float
    blendings: float
    blendings_error: float
    baselines: float
    baselines_error: float

    def __init__(self, file_to_be_read):
        self.number_of_parameters = 7
        with (open(file_to_be_read, 'r') as f):
            lines = f.readlines()
            parameters = lines[0].split(' ')
            errors = lines[1].split(' ')
            self.separation = float(parameters[0])
            self.separation_error = float(errors[0])
            self.mass_ratio = float(parameters[1])
            self.mass_ratio_error = float(errors[1])
            self.u0 = float(parameters[2])
            self.u0_error = float(errors[2])
            self.alpha = float(parameters[3])
            self.alpha_error = float(errors[3])
            self.rho = float(parameters[4])
            self.rho_error = float(errors[4])
            self.tE = float(parameters[5])
            self.tE_error = float(errors[5])
            self.t_star = self.rho * self.tE
            self.t_star_error = np.sqrt(self.tE**2 * self.rho_error**2 + self.rho**2 * self.tE_error**2)
            self.t0 = float(parameters[6])
            self.t0_error = float(errors[6])
            self.chi2 = float(parameters[-1])

            # blends are odd indices starting from 7
            self.blends = np.array(
                [float(parameters[i]) for i in range(self.number_of_parameters, len(parameters) - 1, 2)])
            blends_error = [float(errors[i]) for i in range(self.number_of_parameters, len(errors), 2)]
            # sources are even indices starting from 8
            self.sources = np.array(
                [float(parameters[i]) for i in range(self.number_of_parameters + 1, len(parameters) - 1, 2)])
            sources_error = [float(errors[i]) for i in range(self.number_of_parameters + 1, len(errors), 2)]

            # TODO CHECK THIS
            self.blendings = self.blends / (self.sources + 1.e-12 * self.blends)
            self.blendings_error = blends_error
            self.baselines = -2.5 * np.log10(self.blends + self.sources)
            self.baselines_error = sources_error


@dataclass
class BinaryLensSingleSourceWithParallaxLX:
    """
    LX	Binary_lens_single_source with parallax
    9 parameters
    """
    number_of_parameters: int
    separation: float  # Separation between the lenses in Einstein radii (internally fit in logarithmic (ln) scale)
    separation_error: float
    mass_ratio: float  # Mass ratio of the secondary to the primary lens (internally fit in logarithmic (ln) scale)
    mass_ratio_error: float
    u0: float  # Impact parameter normalized to Einstein angle
    u0_error: float
    alpha: float  # Angle between the source velocity and the vector pointing from the secondary to the primary lens
    alpha_error: float
    rho: float  # Source radius normalized to Einstein angle (internally fit in logarithmic (ln) scale)
    rho_error: float
    tE: float  # Einstein time in days (internally fit in logarithmic (ln) scale)
    tE_error: float
    t_star: float  # Source crossing time in days
    t_star_error: float
    t0: float  # Closest approach time in HJD to the barycenter
    t0_error: float
    piEN: float  # Parallax component along North
    piEN_error: float
    piEE: float  # Parallax component along East
    piEE_error: float
    chi2: float  # Chi2 value for the fit
    blends: float
    sources: float
    blendings: float
    blendings_error: float
    baselines: float
    baselines_error: float

    def __init__(self, file_to_be_read):
        self.number_of_parameters = 9
        with open(file_to_be_read, 'r') as f:
            lines = f.readlines()
            parameters = lines[0].split(' ')
            errors = lines[1].split(' ')
            self.separation = float(parameters[0])
            self.separation_error = float(errors[0])
            self.mass_ratio = float(parameters[1])
            self.mass_ratio_error = float(errors[1])
            self.u0 = float(parameters[2])
            self.u0_error = float(errors[2])
            self.alpha = float(parameters[3])
            self.alpha_error = float(errors[3])
            self.rho = float(parameters[4])
            self.rho_error = float(errors[4])
            self.tE = float(parameters[5])
            self.tE_error = float(errors[5])
            self.t_star = self.rho * self.tE
            self.t_star_error = np.sqrt(self.tE**2 * self.rho_error**2 + self.rho**2 * self.tE_error**2)
            self.t0 = float(parameters[6])
            self.t0_error = float(errors[6])
            self.piEN = float(parameters[7])
            self.piEN_error = float(errors[7])
            self.piEE = float(parameters[8])
            self.piEE_error = float(errors[8])
            self.chi2 = float(parameters[-1])

            # blends are odd indices starting from 9
            self.blends = np.array(
                [float(parameters[i]) for i in range(self.number_of_parameters, len(parameters) - 1, 2)])
            blends_error = [float(errors[i]) for i in range(self.number_of_parameters, len(errors), 2)]
            # sources are even indices starting from 10
            self.sources = np.array(
                [float(parameters[i]) for i in range(self.number_of_parameters + 1, len(parameters) - 1, 2)])
            sources_error = [float(errors[i]) for i in range(self.number_of_parameters + 1, len(errors), 2)]

            # TODO CHECK THIS
            self.blendings = self.blends / (self.sources + 1.e-12 * self.blends)
            self.blendings_error = blends_error
            self.baselines = -2.5 * np.log10(self.blends + self.sources)
            self.baselines_error = sources_error


@dataclass
class BinaryLensSingleSourceWithOrbitalMotionLO:
    """
    LO	Binary_lens_single_source with orbital motion
    12 parameters
    """
    number_of_parameters: int
    separation: float  # Separation between the lenses in Einstein radii (internally fit in logarithmic (ln) scale)
    separation_error: float
    mass_ratio: float  # Mass ratio of the secondary to the primary lens (internally fit in logarithmic (ln) scale)
    mass_ratio_error: float
    u0: float  # Impact parameter normalized to Einstein angle
    u0_error: float
    alpha: float  # Angle between the source velocity and the vector pointing from the secondary to the primary lens
    alpha_error: float
    rho: float  # Source radius normalized to Einstein angle (internally fit in logarithmic (ln) scale)
    rho_error: float
    tE: float  # Einstein time in days (internally fit in logarithmic (ln) scale)
    tE_error: float
    t0: float  # Closest approach time in HJD to the barycenter
    t0_error: float
    t_star: float  # Source crossing time in days
    t_star_error: float
    piEN: float  # Parallax component along North
    piEN_error: float
    piEE: float  # Parallax component along East
    piEE_error: float
    gamma1: float  # Angular velocity parallel to the lens axis
    gamma1_error: float
    gamma2: float  # Angular velocity perpendicular to the lens axis
    gamma2_error: float
    gammaz: float  # Angular velocity along the line of sight
    gammaz_error: float
    chi2: float  # Chi2 value for the fit
    blends: float
    sources: float
    blendings: float
    blendings_error: float
    baselines: float
    baselines_error: float

    def __init__(self, file_to_be_read):
        self.number_of_parameters = 12
        with open(file_to_be_read, 'r') as f:
            lines = f.readlines()
            parameters = lines[0].split(' ')
            errors = lines[1].split(' ')
            self.separation = float(parameters[0])
            self.separation_error = float(errors[0])
            self.mass_ratio = float(parameters[1])
            self.mass_ratio_error = float(errors[1])
            self.u0 = float(parameters[2])
            self.u0_error = float(errors[2])
            self.alpha = float(parameters[3])
            self.alpha_error = float(errors[3])
            self.rho = float(parameters[4])
            self.rho_error = float(errors[4])
            self.tE = float(parameters[5])
            self.tE_error = float(errors[5])
            self.t_star = self.rho * self.tE
            self.t_star_error = np.sqrt(self.tE**2 * self.rho_error**2 + self.rho**2 * self.tE_error**2)
            self.t0 = float(parameters[6])
            self.t0_error = float(errors[6])
            self.piEN = float(parameters[7])
            self.piEN_error = float(errors[7])
            self.piEE = float(parameters[8])
            self.piEE_error = float(errors[8])
            self.gamma1 = float(parameters[9])
            self.gamma1_error = float(errors[9])
            self.gamma2 = float(parameters[10])
            self.gamma2_error = float(errors[10])
            self.gammaz = float(parameters[11])
            self.gammaz_error = float(errors[11])
            self.chi2 = float(parameters[-1])

            self.blends = np.array(
                [float(parameters[i]) for i in range(self.number_of_parameters, len(parameters) - 1, 2)])
            blends_error = [float(errors[i]) for i in range(self.number_of_parameters, len(errors), 2)]
            self.sources = np.array(
                [float(parameters[i]) for i in range(self.number_of_parameters + 1, len(parameters) - 1, 2)])
            sources_error = [float(errors[i]) for i in range(self.number_of_parameters + 1, len(errors), 2)]

            # TODO CHECK THIS
            self.blendings = self.blends / (self.sources + 1.e-12 * self.blends)
            self.blendings_error = blends_error
            self.baselines = -2.5 * np.log10(self.blends + self.sources)
            self.baselines_error = sources_error


if __name__ == "__main__":
    # Example usage
    event_folder = "/Users/stela/Documents/Scripts/orbital_task/RTModel_runs/sample_rtmodel_v2.4/event_0_603_3135"
    event_results = EventResults(event_folder)
    event_results.extract_information_from_nature_file()
    event_results.looking_for_the_names_of_best_model_of_each_category()
    print(event_results)



import pandas as pd


def rtmodel_data_to_pylima(data_folder_path='None', data_output_folder_path='None', event_number='None'):
    # Column names based on the file content
    column_names = ['Mag', 'err', 'HJD_minus_2450000']
    if data_folder_path == 'None':
        data_folder_path = ''
    else:
        data_folder_path = data_folder_path + '/'
    if data_output_folder_path == 'None':
        data_output_folder_path = ''
    else:
        data_output_folder_path = data_output_folder_path + '/'

    # Reading the data into a DataFrame, skipping the first line which is a header comment
    for filepath in ['RomanW146sat1.dat', 'RomanZ087sat2.dat','RomanK213sat3.dat']:
        data = pd.read_csv(f'{data_folder_path}{filepath}',
                           delim_whitespace=True,
                           skiprows=1,
                           names=column_names)

        # Adding 2450000 to the third column
        data['HJD'] = data['HJD_minus_2450000'] + 2450000
        data = data.drop(columns=['HJD_minus_2450000'])

        # Saving the modified data to a new CSV file
        output_txt_path = f'{data_output_folder_path}{event_number:03}_{filepath.split(".")[0]}_pylima.dat'
        data.to_csv(output_txt_path, index=False, sep=' ', header=False)

def rtmodel_data_to_pylima_data_challenge(data_folder_path='None', data_output_folder_path='None', event_number='None'):
    # Column names based on the file content
    column_names = ['Mag', 'err', 'HJD_minus_2450000']
    if data_folder_path == 'None':
        data_folder_path = ''
    else:
        data_folder_path = data_folder_path + '/'
    if data_output_folder_path == 'None':
        data_output_folder_path = ''
    else:
        data_output_folder_path = data_output_folder_path + '/'

    # Reading the data into a DataFrame, skipping the first line which is a header comment
    for filepath in ['RomanW149sat1.dat', 'RomanZ087sat2.dat']:
        data = pd.read_csv(f'{data_folder_path}{filepath}',
                           delim_whitespace=True,
                           skiprows=1,
                           names=column_names)

        # Adding 2450000 to the third column
        data['HJD'] = data['HJD_minus_2450000'] + 2450000
        data = data.drop(columns=['HJD_minus_2450000'])

        # Saving the modified data to a new CSV file
        output_txt_path = f'{data_output_folder_path}{event_number:03}_{filepath.split(".")[0]}_pylima.dat'
        data.to_csv(output_txt_path, index=False, sep=' ', header=False)
if __name__ == '__main__':
    rtmodel_data_to_pylima()
    print('Data files created successfully.')



from typing import Dict
from dataclasses import dataclass

import pandas as pd


@dataclass
class Event:
    name: str
    l_deg: float
    b_deg: float
    ra_deg: float
    dec_deg: float
    lightcurves: Dict[str, pd.DataFrame]

    @classmethod
    def from_name(cls, name: str, df):
        df_event = df[df["name"] == name]

        if df_event.empty:
            raise ValueError(f"Event {name} not found")

        meta = df_event.iloc[0]

        lightcurves = {}
        for filters, df_filters in df_event.groupby("filt"):
            # keep only relevant columns (and copy to avoid view issues)
            lightcurves[filters] = df_filters[
                [
                    "bjd",
                    "mag",
                    "mag_err",
                    "extinction",
                    "obs_x",
                    "obs_y",
                    "obs_z",
                    "saturation_flag",
                ]
            ].copy()

        return cls(
            name=name,
            l_deg=meta["l_deg"],
            b_deg=meta["b_deg"],
            ra_deg=meta["ra_deg"],
            dec_deg=meta["dec_deg"],
            lightcurves=lightcurves,
        )

    @property
    def simple_lightcurve(self) -> Dict[str, pd.DataFrame]:
        return {
            filters: df_event[["bjd", "mag", "mag_err"]].copy()
            for filters, df_event in self.lightcurves.items()
        }


if __name__ == "__main__":
    # Example usage
    # Read parquet file as a pandas dataframe
    df = pd.read_parquet(
        "../../../../data_hugging_face/RMDC26_Beginner_Tier_test.parquet"
    )
    df["bjd"] = df["bjd"] - 2_450_000  # adjust if you would like to

    # Load the event you want:
    event = Event.from_name("RMDC26_000001", df)
    print("Event name: ", event.name)
    print("Event l_deg: ", event.l_deg)
    print("Event b_deg: ", event.b_deg)
    print("Event ra_deg: ", event.ra_deg)
    print("Event dec_deg: ", event.dec_deg)

    complete_lightcurve = event.lightcurves["F146"]  # pandas dataframe
    print("Part of complete lightcurve: \n", complete_lightcurve.head(3))
    simple_lightcurve = event.simple_lightcurve["F146"]  # pandas dataframe
    print("Part of simple lightcurve: \n", simple_lightcurve.head(3))

    print()



#from gridfit import *
#import sys
#import tdqm
import numpy as np
from jasmine.initial_condition_grid_search.gridfit import *


def main(event_path):
    a1_list = [0.33,0.4534,0.2517]
    dataset_list = ['RomanW146sat1.dat','RomanZ087sat2.dat','RomanK213sat3.dat']
    #s = '0.37 0.40 0.43 0.46 0.49 0.52 0.56 0.60 0.65 0.70 0.75 0.80 0.84 0.88 0.91 0.93 0.95 0.965 0.98 0.99 1.00 1.01 1.02 1.035 1.05 1.07 1.09 1.12 1.15 1.18 1.22 1.26 1.30 1.34 1.38 1.42 1.46 1.50 1.55 1.60 1.66 1.72 1.79 1.86 1.94 2.02 2.1 2.2 2.3 2.4 2.5 2.62 2.75'
    #s = s.split(' ')
    #s_grid = list(map(float,s))
    eps1 = '1.e-6 3.16e-6 1.e-5 3.16e-5 1.e-4 3.16e-4 1.e-3 3.16e-3 1.e-2 3.16e-2 1.e-1 3.16e-1'
    eps1 = eps1.split(' ')
    eps1_grid = np.array(list(map(float,eps1)))
    q_grid = 1/((1/eps1_grid)-1)
    tstar=0.03

    d_alpha = 0.005
    n_alpha = 20
    alpha_frac = 0.05

    s_frac = 0.01
    n_sep = 10

    #alpha_grid = np.linspace(-3.15,3.15,316)
    dir6 = '/Users/jmbrashe/VBBOrbital/NEWGULLS/6_events_for_testing'
    event_name=event_path
    event_path = f'{dir6}/{event_path}'
    satdir = '/Users/jmbrashe/VBBOrbital/NEWGULLS/6_events_for_testing/levmar_runs_good_plx/satellitedir'
    run_event_rtm_grid(event_path=event_path, dataset_list=dataset_list,grid_q=q_grid,tstar=tstar,d_alpha=d_alpha, n_alpha=n_alpha, alpha_frac=alpha_frac,
                       s_frac=s_frac, n_sep=n_sep,a1_list=a1_list, pspl_thresh=0, satellitedir=satdir,processors=1)
if __name__ == "__main__":
    #dir6 = '/Users/jmbrashe/VBBOrbital/NEWGULLS/6_events_for_testing/levmar_runs_redo'
    event = 'event_0_5_606'
    #with multiprocessing.Pool(1) as p:
    #    p.map(main,events)
    main(event)









"""
Run this code if you want a quick summary of chi2, q, s, and piE values for the best models of each event.
"""

import glob
import pandas as pd
import os

import jasmine.classes_and_files_reader.mass_ratio_and_separation_getter as q_s_getter
import jasmine.classes_and_files_reader.parallax_getter as pie_getter


def chi2_getter(filepath_):
    """
    This function reads the chi2 file and returns the chi2 value
    :param filepath_:
    :return: chi2
    """
    with open(filepath_, 'r') as f:
        lines = f.readlines()
        chi2 = float(lines[0].split(' ')[-1])
    return chi2


def models_per_chi2_rank(folder_path_):
    """
    This function rank the models per chi2 value and create a chi2_top1_of_each_binary_lens_model.csv file
    """
    models = glob.glob(folder_path_ + '/Models/*txt')
    models_names = [model.replace(folder_path_, '').replace('/Models/', '').replace('.txt', '') for model in models]
    chi2_values = []
    for model in models:
        chi2_values.append(chi2_getter(model))
    chi2_df = pd.DataFrame({'model': models_names, 'chi2': chi2_values})
    chi2_df = chi2_df.sort_values(by='chi2')
    chi2_df.to_csv(folder_path_ + '/Models/chi2_rank.csv', index=False)
    only_binary_lenses = chi2_df[chi2_df['model'].str.contains('L')]
    only_binary_lenses.to_csv(folder_path_ + '/Models/chi2_rank_binary_lenses.csv', index=False)
    only_ls = chi2_df[chi2_df['model'].str.contains('LS')]
    only_lx = chi2_df[chi2_df['model'].str.contains('LX')]
    only_lo = chi2_df[chi2_df['model'].str.contains('LO')]
    # top_1_of_each = pd.concat([only_ls.iloc[[0]], only_lx.iloc[[0]], only_lo.iloc[[0]]])

    # Collect the top 1 row from each type if available
    dfs_to_concat = [df.iloc[[0]] for df in [only_ls, only_lx, only_lo] if not df.empty]

    if dfs_to_concat:
        top_1_of_each = pd.concat(dfs_to_concat, ignore_index=True)
        # Save the concatenated result to CSV
        top_1_of_each.to_csv(folder_path_ + '/Models/chi2_top1_of_each_binary_lens_model.csv', index=False)
        return True
    else:
        event_name = folder_path_.split('/')[-1]
        print(f"No valid models to concatenate. "
              f"Skipping creation of 'chi2_top1_of_each_binary_lens_model.csv' for {event_name}.")
        return False


def main(general_path_, list_of_events_, parallax=True, mass_ratio_and_separation=True, type_of_event_='top10_piE'):
    """
    This function is the main function that calls the other functions to get the summary of chi2, q, s, and piE values
    :param general_path_:
    :param list_of_events_:
    :param parallax:
    :param mass_ratio_and_separation:
    :param type_of_event_:
    :return:
    """
    skipped_event = []
    for event in list_of_events_:
        folder_path_ = f'{general_path_}/{event}'
        models_folder = os.path.join(folder_path_, 'Models')
        if not os.path.exists(models_folder):
            print(f"Folder 'Models' doesn't exist for event {event}. Skipping this event.")
            skipped_event.append(event)
            continue
        are_there_binary_solutions = models_per_chi2_rank(folder_path_)
        if mass_ratio_and_separation:
            q_s_getter.get_summary_of_q_s_chi2_per_event(folder_path_, are_there_binary_solutions)
        if parallax:
            pie_getter.get_summary_of_parallax_per_event(folder_path_, are_there_binary_solutions)
    # Remove skipped events from the list
    print('Skipped events:', skipped_event)
    list_of_events_ = [event for event in list_of_events_ if event not in skipped_event]
    if mass_ratio_and_separation:
        q_s_getter.event_summary_q_s_wrapper(general_path_, list_of_events_, type_of_event_)
    if parallax:
        pie_getter.event_summary_parallax_wrapper(general_path_, list_of_events_, type_of_event_)


def list_of_events_from_sample_df(runs_path_):
    """
    This function reads the sample file and returns a list of events
    :param runs_path_:
    :return:
    """
    general_path_ = runs_path_.split('/RTModel_runs')[0] + '/data/gulls_orbital_motion_extracted'
    master_file_path = f'{general_path_}/OMPLDG_croin_cassan.sample.csv'
    sample_df = pd.read_csv(master_file_path)
    list_of_events_ = []
    for _, row in sample_df.iterrows():
        subrun = row['SubRun']
        event_id = row['EventID']
        field = row['Field']
        event_name = f'event_{subrun}_{field}_{event_id}'
        list_of_events_.append(event_name)
    return list_of_events_


if __name__ == '__main__':
    # general_path_outputs = '/discover/nobackup/sishitan/orbital_task/RTModel_runs/top10_piE'
    # list_of_events = ['event_0_762_407',
    #                   'event_0_876_1031',
    #                   'event_1_793_3191',
    #                   'event_0_42_270',
    #                   'event_0_922_1199',
    #                   'event_0_672_2455',
    #                   'event_0_992_224',
    #                   'event_0_42_2848',
    #                   'event_1_755_564',
    #                   'event_0_798_371']

    #
    # general_path_outputs = '/Users/stela/Documents/Scripts/orbital_task/RTModel_runs/top10_piE'
    # list_of_events = ['event_0_42_270',]
    # # 'event_0_1000_1445',

    # type_of_event = 'sample'
    # type_of_event = '6_events_for_testing'
    # type_of_event = 'icgs_levmar'
    # type_of_event = 'sample_rtmodel_v2.4'
    type_of_event = 'sample_rtmodel_v3.1'
    # runs_path = f'/Users/stela/Documents/Scripts/orbital_task/RTModel_runs/{type_of_event}'
    runs_path = f'/discover/nobackup/sishitan/orbital_task/RTModel_runs/{type_of_event}'
    list_of_events = list_of_events_from_sample_df(runs_path)
    # list_of_events = ['event_0_128_2350',
    #                   'event_0_167_1273',
    #                   'event_0_672_793',
    #                   'event_0_715_873',
    #                   'event_0_874_19',
    #                   'event_0_952_2841']
    main(runs_path, list_of_events, parallax=True, mass_ratio_and_separation=True, type_of_event_=type_of_event)



import VBMicrolensing
import pandas as pd
import numpy as np
import RTModel
import os
import shutil
import matplotlib.pyplot as plt
import time

import glob
from joblib import Parallel, delayed
from jasmine.files_organizer import ra_and_dec_conversions as radec
from pyLIMA import event
from pyLIMA import telescopes
from pyLIMA.models import PSPL_model
from pyLIMA.fits import DE_fit, TRF_fit


def minimize_linear_pars(y, err, x):
    """
    Exact solution to minimize linear eq chi2
    """
    alpha1 = np.sum(x / err ** 2)
    alpha2 = np.sum(1 / err ** 2)
    alpha3 = np.sum(y / err ** 2)
    alpha4 = np.sum((x / err) ** 2)
    alpha5 = np.sum(x * y / err ** 2)
    a = (alpha5 * alpha2 - alpha1 * alpha3) / (alpha4 * alpha2 - alpha1 ** 2)
    b = (alpha4 * alpha3 - alpha5 * alpha1) / (alpha4 * alpha2 - alpha1 ** 2)
    return a, b


def get_chi2(magnitude_list, mag_data_list, error_list, ndatasets):
    """calculate chi2 for model on each dataset provided"""
    chi2_list = []
    for i in range(ndatasets):
        res = (magnitude_list[i] - mag_data_list[i]) / error_list[i]
        chi2 = np.sum(np.power(res, 2))
        chi2_list.append(chi2)
    chi2_sum = np.sum(chi2_list)

    return chi2_list, chi2_sum


def get_residuals(magnitude_list, mag_data_list, error_list, ndatasets):
    """calculate residuals for model on each dataset provided"""
    residue_list = []
    for i in range(ndatasets):
        res = (magnitude_list[i] - mag_data_list[i]) / error_list[i]
        residue_list.append(res)
    residues = np.concatenate(residue_list)
    return residues


def calc_pspl_chi2(pars, args):
    """ Objective function to optimize for pspl_fit() """
    magnitude_list = []
    source_flux_list = []
    blend_flux_list = []
    data_list = args[0]
    ndatasets = args[1]
    VBMInstance = args[2]
    # logger = args[3]
    pars_log = [np.log(pars[0]), np.log(pars[1]), pars[2]]
    # logger.info(pars)
    fluxdata_list = []
    fluxerr_list = []
    flux_list = []
    for i in range(ndatasets):
        data = data_list[i]
        magnifications, y1, y2 = VBMInstance.PSPLLightCurve(pars_log, data[:, -1])
        meas_flux = 10 ** (-0.4 * data[:, 0])
        meas_flux_err = 0.4 * np.log(10) * 10 ** (-0.4 * data[:, 0]) * data[:, 1]
        fluxdata_list.append(meas_flux)
        fluxerr_list.append(meas_flux_err)
        source_flux, blend_flux = minimize_linear_pars(meas_flux, meas_flux_err, magnifications)
        sim_flux = np.array(magnifications) * source_flux + blend_flux
        flux_list.append(sim_flux)
        source_flux_list.append(source_flux)
        blend_flux_list.append(blend_flux)
        sim_magnitudes = -2.5 * np.log10(sim_flux)
        magnitude_list.append(sim_magnitudes)
    magdata_list = []
    magerr_list = []
    for i in range(ndatasets):
        magdata_list.append(data_list[i][:, 0])
        magerr_list.append(data_list[i][:, 1])
    chi2_list, chi2 = get_chi2(magnitude_list, magdata_list, magerr_list, ndatasets)
    chi2_listf, chi2f = get_chi2(flux_list, fluxdata_list, fluxerr_list, ndatasets)
    print('VBM PSPL chi2 with mag ', chi2)
    print('VBM PSPL chi2 with flux ', chi2f)
    print('Source Fluxes ', source_flux_list)
    print('Blend Fluxes ', blend_flux_list)
    return chi2


def calculate_magnifications(pars, a1_list, data_list, ndatasets, VBMInstance, parallax=False):
    """Calculate LC magnitudes in each passband provided for 2L1S"""
    magnitude_list = []
    flux_list = []
    fluxdata_list = []
    fluxerr_list = []
    source_flux_list = []
    blend_flux_list = []
    # loop over bands
    # print(pars)
    for i in range(ndatasets):
        # must update a1 each time
        VBMInstance.a1 = a1_list[i]
        data = data_list[i]
        if parallax:
            # print(pars)
            # set satellite for each band if PLX
            VBMInstance.satellite = i + 1
            # print(VBMInstance.satellite)
            magnifications, y1, y2 = VBMInstance.BinaryLightCurveParallax(pars, data[:, -1])
        else:
            magnifications, y1, y2 = VBMInstance.BinaryLightCurve(pars, data[:, -1])
        # get the fluxes from measured lightcurve magnitudes
        meas_flux = 10 ** (-0.4 * data[:, 0])
        meas_flux_err = 0.4 * np.log(10) * 10 ** (-0.4 * data[:, 0]) * data[:, 1]
        fluxdata_list.append(meas_flux)
        fluxerr_list.append(meas_flux_err)
        # Fix so minimize_linear_parameters uses flux_err not mag_err
        # finds optimal linear source + blend flux parameters
        source_flux, blend_flux = minimize_linear_pars(meas_flux, meas_flux_err, magnifications)
        sim_flux = np.array(magnifications) * source_flux + blend_flux  # model lightcurve fluxes
        source_flux_list.append(source_flux)
        blend_flux_list.append(blend_flux)
        sim_magnitudes = -2.5 * np.log10(sim_flux)  # model lightcurve magnitudes
        magnitude_list.append(sim_magnitudes)
        flux_list.append(sim_flux)

        #
    return magnitude_list, flux_list, fluxdata_list, fluxerr_list, source_flux_list, blend_flux_list


def calculate_magnifications_plots(pars, a1_list, data_list, data_mag, ndatasets, VBMInstance, parallax=False):
    """Calculate LC magnitudes in each passband provided for 2L1S"""
    magnitude_list = []
    flux_list = []
    fluxdata_list = []
    fluxerr_list = []
    source_flux_list = []
    blend_flux_list = []
    for i in range(ndatasets):
        VBMInstance.a1 = a1_list[i]
        data = data_list[i]
        if parallax == False:
            magnifications, y1, y2 = VBMInstance.BinaryLightCurve(pars, data[:, -1])
        else:
            magnifications, y1, y2, sorb = VBMInstance.BinaryLightCurveOrbital(pars, data[:, -1])
        meas_flux = 10 ** (-0.4 * data[:, 0])
        meas_flux_err = 0.4 * np.log(10) * 10 ** (-0.4 * data[:, 0]) * data[:, 1]
        fluxdata_list.append(meas_flux)
        fluxerr_list.append(meas_flux_err)
        # Fix so minimize_linear_parameters uses flux_err not mag_err
        source_flux, blend_flux = minimize_linear_pars(meas_flux, meas_flux_err, magnifications)
        magnifications, y1, y2 = VBMInstance.BinaryLightCurve(pars, data_mag)
        sim_flux = np.array(magnifications) * source_flux + blend_flux
        source_flux_list.append(source_flux)
        blend_flux_list.append(blend_flux)
        sim_magnitudes = -2.5 * np.log10(sim_flux)
        magnitude_list.append(sim_magnitudes)
        flux_list.append(sim_flux)
    return magnitude_list, flux_list, fluxdata_list, fluxerr_list, source_flux_list, blend_flux_list


def evaluate_model(psplpars, fsblpars, a1_list, data_list, VBMInstance, pspl_chi2, parallax=False):
    """ Calculate chi2 for one model in the grid
        fsblpars = [s,q,alpha,tstar]
        psplpars = follows VBMicrolensing format
        data_list list with a dataframe for each band in RTModel data format
        ndatasets = number of datasets
    """
    # blend_flux_list = []
    # source_flux_list = []
    ndatasets = len(data_list)
    # create parameter array for VBM.BinaryLightCurve()
    if parallax:
        pars = [np.log(fsblpars[0]), np.log(fsblpars[1]), psplpars[0], fsblpars[2], np.log(fsblpars[3] / psplpars[1]),
                np.log(psplpars[1]), psplpars[2], psplpars[3], psplpars[4]]
    else:
        pars = [np.log(fsblpars[0]), np.log(fsblpars[1]), psplpars[0], fsblpars[2],
                np.log(fsblpars[3] / psplpars[1]), np.log(psplpars[1]), psplpars[2]]
    # Now go to calculate_magnifications
    # This function will use VBM to get mags, then calculate source + blend fluxes.
    magnitude_list, flux_list, fluxdata_list, fluxerr_list, source_flux_list, blend_flux_list = calculate_magnifications(
        pars, a1_list, data_list, ndatasets, VBMInstance, parallax=parallax)
    # Just the way I passed measured magnitudes to get_chi2(). Maybe I should change it because it's really terrible.
    magdata_list = []
    magerr_list = []
    for i in range(ndatasets):
        magdata_list.append(data_list[i][:, 0])
        magerr_list.append(data_list[i][:, 1])
    # print(source_flux_list)

    chi2_list, chi2_sum = get_chi2(flux_list, fluxdata_list, fluxerr_list, ndatasets)
    # return this, it is one line of the grid output file from grid_fit()
    return_list = np.concatenate((pars, source_flux_list, blend_flux_list, chi2_list, [chi2_sum, chi2_sum - pspl_chi2]))

    # fig, ax = plt.subplots()
    # ax.errorbar(data_list[0][:,-1],data_list[0][:,0],yerr = data_list[0][:,1], marker='.', markersize=0.75,
    #            linestyle=' ', label='W146')
    # ax.plot(data_list[0][:,-1], magnitude_list[0], zorder=10, label='2L1S Fit', color='black')

    # ax.annotate(f'W146 chi2 = {chi2_list[0]}\nTotal chi2 = {chi2_sum}',xy=(0.8,0.9))
    # ax.invert_yaxis()
    # ax.set_xlim(10000, 10050)
    # ax.legend()
    # plt.show()
    # print(' ')
    return return_list


def grid_fit_pass_meshgrid(event_path, dataset_list, pspl_pars, s, q, alpha, tstar, a1_list, pspl_chi2):
    """
        Fit grid of models to determine initial conditions
        pass preconstructed meshgrid to the grid fit
    """
    data_list = []
    for i in range(len(dataset_list)):
        data_list.append(np.loadtxt(f'{event_path}/Data/{dataset_list[i]}'))
    VBMInstance = VBMicrolensing.VBMicrolensing()
    VBMInstance.RelTol = 1e-03
    VBMInstance.Tol = 1e-03
    # print(s.shape[0])
    grid_results = np.zeros(shape=(s.shape[0], 9 + 3 * len(a1_list)))
    print(f'PSPL pars: {pspl_pars[0]} {pspl_pars[1]} {pspl_pars[2]}')
    print(f'Checking {s.shape[0]} Models on grid!')
    for i in range(s.shape[0]):
        fsblpars = [s[i], q[i], alpha[i], tstar]
        print(f'{i} {s[i]} {q[i]} {alpha[i]} {tstar}')
        output = evaluate_model(pspl_pars, fsblpars, a1_list, data_list, VBMInstance, pspl_chi2)
        grid_results[i, :] = output
        # if i%5000==0: print(f'{i} Models checked')
    print('Done checking models!')
    return grid_results


def make_grid(grid_s, grid_q, grid_alpha, use_croin=False):
    if use_croin:
        print('Implementation in progress')
    else:
        s, q, alpha = np.meshgrid(grid_s, grid_q, grid_alpha)
        s = s.flatten()
        q = q.flatten()
        alpha = alpha.flatten()
    return s, q, alpha


# Functions for compressing 3 nested loops into one and recovering the three indices
def CALCULATE_INDEX(i, j, k, ni, nj, nk):
    IND = i * nj * nk + j * nk + k
    return IND


def RECOVER_INDICES(IND, ni, nj, nk):
    i = np.int64(IND / (nj * nk))
    left = IND - (i * nj * nk)
    j = np.int64(left / nk)
    k = left - j * nk
    return i, j, k


def create_random_indices(event_path, nmodels, ncores):
    '''
    Divide nmodels on the grid into ncore batches to run.
    Randomize index order so one job doesn't get stuck on all the long ones
    '''
    index = np.arange(0, nmodels)
    nper_job = int(index.shape[0] / ncores)
    nper = nper_job * np.ones(ncores)
    left = index.shape[0] - ncores * nper_job
    for i in range(left):
        nper[i] += 1
    randomized_index = np.random.choice(index, nmodels, replace=False)
    start_ind = np.zeros(ncores, dtype=int)
    start_ind[1:ncores] = np.cumsum(nper, dtype=int)[0:ncores - 1]
    end_ind = np.cumsum(nper, dtype=int)[0:ncores]
    for i in range(ncores):
        sub_indices = randomized_index[start_ind[i]:end_ind[i]]
        #np.save(f'{event_path}/Data/sub_index_list_{i}.npy', sub_indices)
    return randomized_index, start_ind, end_ind


def evaluation_loop(common_parameters, index_list, iteration):
    # unpack values. This is why we need a class instead
    event_path = common_parameters[0]
    satellitedir = common_parameters[1]
    grid_q = common_parameters[2]
    grid_s = common_parameters[3]
    grid_alpha = common_parameters[4]
    tstar = common_parameters[5]
    pspl_pars = common_parameters[6]
    a1_list = common_parameters[7]
    data_list = common_parameters[8]
    pspl_chi2 = common_parameters[9]
    parallax = common_parameters[10]
    del common_parameters  # get that outta here

    VBMInstance = VBMicrolensing.VBMicrolensing()
    VBMInstance.parallaxsystem = 1  # :( :( :( :(
    # VBMInstance.t0_par_fixed = 1
    # VBMInstance.t0_par = 10025.111593759
    # if Parallax set coordinates and read in satellite tables
    if parallax:
        VBMInstance.SetObjectCoordinates(f'{event_path}/Data/event.coordinates', satellitedir)
    VBMInstance.Tol = 1.0e-02
    VBMInstance.RelTol = 1.0e-03  # leave at 0 for ICGS
    VBMInstance.minannuli = 2

    with open(f'{event_path}/Data/grid_fit{iteration}.txt', 'w') as grid_file:
        for index in index_list:
            grid_calculation_time_start = time.time() # Adding time track for each model evaluation/calculation
            i, j, k = RECOVER_INDICES(index, grid_q.shape[0], grid_s.shape[0], grid_alpha.shape[0])
            fsblpars = [grid_s[j], grid_q[i], grid_alpha[k], tstar]
            # evaluate model is where chi2 is calculated.
            output = evaluate_model(pspl_pars, fsblpars, a1_list, data_list, VBMInstance, pspl_chi2, parallax=parallax)
            grid_calculation_time_end = time.time()
            grid_calculation_time = grid_calculation_time_end - grid_calculation_time_start
            grid_file.write(f'{index} {output[-2]} {grid_calculation_time}\n')


def grid_fit_parallelized(event_path, dataset_list, pspl_pars, grid_s, grid_q, grid_alpha, tstar, a1_list, pspl_chi2,
                          parallax=False, satellitedir=None, use_croin=False, nprocessors=2):
    """
        Fit grid of models to determine initial conditions
    """
    # read datasets from /Data
    data_list = []
    for i in range(len(dataset_list)):
        data_list.append(np.loadtxt(f'{event_path}/Data/{dataset_list[i]}'))
    # create VBM instance

    # I was experimenting with fixing the t0_par to exactly what pyLIMA uses, didn't fix anything.
    # create parameter meshgrid
    # s,q,alpha = make_grid(grid_s,grid_q, grid_alpha, use_croin)
    # print(s.shape[0])
    n_models = grid_q.shape[0] * grid_s.shape[0] * grid_alpha.shape[0]
    randomized_index, start, end = create_random_indices(event_path, n_models, nprocessors)

    # Initialize results array, with size depending on if static
    # N parameters + source+blend fluxes + 3 bands chi2 + total chi2 + delta chi2 = 7/9 + 2+3*len(al_list)
    # calc_pspl_chi2(pspl_pars,[data_list,3,VBMInstance])

    print(f'Checking {n_models} Models on grid with {nprocessors} cores!')
    # loop over each model in the grid and call evaluate model
    # This is where the parallelization must be implemented. So write a function here
    # INDICES for grids
    common_parameters = [event_path, satellitedir, grid_q, grid_s, grid_alpha, tstar, pspl_pars, a1_list,
                         data_list, pspl_chi2, parallax]
    Parallel(n_jobs=nprocessors)(
        delayed(evaluation_loop)(common_parameters, randomized_index[start[iteration]:end[iteration]], iteration) for
        iteration in range(nprocessors))
    print('Done checking models!')
    return None


def grid_fit(event_path, dataset_list, pspl_pars, grid_s, grid_q, grid_alpha, tstar, a1_list, pspl_chi2, parallax=False,
             satellitedir=None, use_croin=False):
    """
        Fit grid of models to determine initial conditions
    """
    # read datasets from /Data
    data_list = []
    for i in range(len(dataset_list)):
        data_list.append(np.loadtxt(f'{event_path}/Data/{dataset_list[i]}'))
    # create VBM instance
    VBMInstance = VBMicrolensing.VBMicrolensing()
    VBMInstance.parallaxsystem = 1  # :( :( :( :(
    # VBMInstance.t0_par_fixed = 1
    # VBMInstance.t0_par = 10025.111593759
    # if Parallax set coordinates and read in satellite tables
    if parallax:
        VBMInstance.SetObjectCoordinates(f'{event_path}/Data/event.coordinates', satellitedir)
        # VBMInstance.SetObjectCoordinates('17:55:35.00561287 -30:12:38.19570995')
    VBMInstance.Tol = 1e-02
    VBMInstance.RelTol = 0  # leave at 0 for ICGS
    VBMInstance.minannuli = 2
    # I was experimenting with fixing the t0_par to exactly what pyLIMA uses, didn't fix anything.
    # create parameter meshgrid
    # s,q,alpha = make_grid(grid_s,grid_q, grid_alpha, use_croin)
    # print(s.shape[0])
    n_models = grid_q.shape[0] * grid_s.shape[0] * grid_alpha.shape[0]
    # Initialize results array, with size depending on if static
    # N parameters + source+blend fluxes + 3 bands chi2 + total chi2 + delta chi2 = 7/9 + 2+3*len(al_list)
    # calc_pspl_chi2(pspl_pars,[data_list,3,VBMInstance])
    if parallax:
        # grid_results = np.zeros(shape=(n_models, 11 + 3 * len(a1_list)))
        print(f'PSPL + PLX pars: {pspl_pars[0]} {pspl_pars[1]} {pspl_pars[2]} {pspl_pars[3]} {pspl_pars[4]}')
    else:
        # grid_results = np.zeros(shape=(n_models, 9 + 3 * len(a1_list)))
        print(f'PSPL pars: {pspl_pars[0]} {pspl_pars[1]} {pspl_pars[2]}')
    print(f'Checking {n_models} Models on grid!')
    # loop over each model in the grid and call evaluate model
    # This is where the parallelization must be implemented. So write a function here
    # INDICES for grids
    iter = 0
    with open(f'{event_path}/Data/grid_fit{iter}.txt', 'w') as grid_file:
        for index in range(n_models):
            grid_calculation_time_start = time.time() # Adding time track for each model evaluation/calculation
            i, j, k = RECOVER_INDICES(index, grid_q.shape[0], grid_s.shape[0], grid_alpha.shape[0])
            fsblpars = [grid_s[j], grid_q[i], grid_alpha[k], tstar]
            # evaluate model is where chi2 is calculated.
            output = evaluate_model(pspl_pars, fsblpars, a1_list, data_list, VBMInstance, pspl_chi2, parallax=parallax)
            grid_calculation_time_end = time.time()
            grid_calculation_time = grid_calculation_time_end - grid_calculation_time_start
            grid_file.write(f'{index} {output[-2]} {grid_calculation_time}\n')
            # grid_results[i,:] = output
        # print(f'{index} {s[i]} {q[i]} {alpha[i]} {tstar} {pspl_pars[0]} {pspl_pars[1]} {pspl_pars[2]} {output[-2]}') # print this to see chi2 of each model.
        # if i%10==0: print(f'{i} Models checked')
    print('Done checking models!')
    return None


# def evaluate_multi()

def pspl_fit_pyLIMA(event_path, dataset_list):
    data_list = []
    for i in range(len(dataset_list)):
        data_list.append(np.loadtxt(f'{event_path}/Data/{dataset_list[i]}'))
        data_list[i] = data_list[i][:, [2, 0, 1]]
        data_list[i][:, 0] += 2450000  # add +2_450_000 for pyLIMA
    print('Using pyLIMA DE')

    file_path = f'{event_path}/Data/event.coordinates'
    event_coordinates = pd.read_csv(file_path, delim_whitespace=True, header=None, names=['RA_HMS', 'Dec_DMS'])
    # Convert RA and Dec to decimal degrees
    ra_deg = radec.ra_hms_to_deg(event_coordinates['RA_HMS'].iloc[0])
    dec_deg = radec.dec_dms_to_deg(event_coordinates['Dec_DMS'].iloc[0])

    your_event = event.Event(ra=ra_deg, dec=dec_deg)
    your_event.name = f'PSPL Fit'
    telescope_1 = telescopes.Telescope(name='RomanW146',
                                       camera_filter='W146',
                                       lightcurve=data_list[0],
                                       lightcurve_names=['time', 'mag', 'err_mag'],
                                       lightcurve_units=['JD', 'mag', 'mag'])
    telescope_2 = telescopes.Telescope(name='RomanZ087',
                                       camera_filter='Z087',
                                       lightcurve=data_list[1],
                                       lightcurve_names=['time', 'mag', 'err_mag'],
                                       lightcurve_units=['JD', 'mag', 'mag'])
    telescope_3 = telescopes.Telescope(name='RomanK213',
                                       camera_filter='K213',
                                       lightcurve=data_list[2],
                                       lightcurve_names=['time', 'mag', 'err_mag'],
                                       lightcurve_units=['JD', 'mag', 'mag'])
    your_event.telescopes.append(telescope_1)
    your_event.telescopes.append(telescope_2)
    your_event.telescopes.append(telescope_3)
    your_event.find_survey('RomanW146')
    your_event.check_event()
    # print(telescope_1.bad_data)
    pspl = PSPL_model.PSPLmodel(your_event, parallax=['None', 0.0])

    diffev_fit = DE_fit.DEfit(pspl, loss_function='chi2')
    diffev_fit.fit_parameters['u0'][1] = [0., 5.]
    diffev_fit.fit_parameters['tE'][1] = [0.1, 200.0]
    diffev_fit.fit()
    # Do gradient fit to fine tune solution
    gradient_fit = TRF_fit.TRFfit(pspl, loss_function='chi2')
    gradient_fit.fit_parameters['u0'][1] = [0, 5.]  # PyLima has a short limit for u0
    gradient_fit.fit_parameters['tE'][1] = [0.1, 200.0]

    gradient_fit.model_parameters_guess = diffev_fit.fit_results['best_model'][0:3]
    gradient_fit.fit()

    results = gradient_fit.fit_results['best_model']
    chi2 = gradient_fit.fit_results['chi2']
    return results, chi2


def psplPLX_fit_pyLIMA(event_path, dataset_list, satellitedir):
    data_list = []
    for i in range(len(dataset_list)):
        data_list.append(np.loadtxt(f'{event_path}/Data/{dataset_list[i]}'))
        data_list[i] = data_list[i][:, [2, 0, 1]]
        data_list[i][:, 0] += 2450000  # add +2_450_000 for pyLIMA
    print('Using pyLIMA DE')

    file_path = f'{event_path}/Data/event.coordinates'
    event_coordinates = pd.read_csv(file_path, delim_whitespace=True, header=None, names=['RA_HMS', 'Dec_DMS'])
    # Convert RA and Dec to decimal degrees
    ra_deg = radec.ra_hms_to_deg(event_coordinates['RA_HMS'].iloc[0])
    dec_deg = radec.dec_dms_to_deg(event_coordinates['Dec_DMS'].iloc[0])

    your_event = event.Event(ra=ra_deg, dec=dec_deg)
    your_event.name = f'PSPL + PLX Fit'
    telescope_1 = telescopes.Telescope(name='RomanW146',
                                       camera_filter='W146',
                                       lightcurve=data_list[0],
                                       lightcurve_names=['time', 'mag', 'err_mag'],
                                       lightcurve_units=['JD', 'mag', 'mag'])
    telescope_2 = telescopes.Telescope(name='RomanZ087',
                                       camera_filter='Z087',
                                       lightcurve=data_list[1],
                                       lightcurve_names=['time', 'mag', 'err_mag'],
                                       lightcurve_units=['JD', 'mag', 'mag'])
    telescope_3 = telescopes.Telescope(name='RomanK213',
                                       camera_filter='K213',
                                       lightcurve=data_list[2],
                                       lightcurve_names=['time', 'mag', 'err_mag'],
                                       lightcurve_units=['JD', 'mag', 'mag'])

    ### IMPORTANT: Tell the code that Roman is in space!
    # read in ephemerides files
    names = ['time', 'RA', 'Dec', 'dist', 'deldot']
    t1_ephem = pd.read_csv(f'{satellitedir}/satellite1.txt', names=names, comment='$', sep='\s+')
    t2_ephem = pd.read_csv(f'{satellitedir}/satellite2.txt', names=names, comment='$', sep='\s+')
    t3_ephem = pd.read_csv(f'{satellitedir}/satellite3.txt', names=names, comment='$', sep='\s+')

    telescope_1.location = 'Space'
    telescope_1.spacecraft_name = 'Roman'
    telescope_1.spacecraft_positions['photometry'] = t1_ephem[['time', 'RA', 'Dec', 'dist']].values
    telescope_2.location = 'Space'
    telescope_2.spacecraft_name = 'Roman'
    telescope_2.spacecraft_positions['photometry'] = t2_ephem[['time', 'RA', 'Dec', 'dist']].values
    telescope_3.location = 'Space'
    telescope_3.spacecraft_name = 'Roman'
    telescope_3.spacecraft_positions['photometry'] = t3_ephem[['time', 'RA', 'Dec', 'dist']].values

    your_event.telescopes.append(telescope_1)
    your_event.telescopes.append(telescope_2)
    your_event.telescopes.append(telescope_3)
    your_event.find_survey('RomanW146')
    your_event.check_event()
    # print(telescope_1.bad_data)

    # print(telescope_1.bad_data)
    pspl_noplx = PSPL_model.PSPLmodel(your_event, parallax=['None', 0.0])

    diffev_fit = DE_fit.DEfit(pspl_noplx, loss_function='chi2')
    diffev_fit.fit_parameters['u0'][1] = [0., 5.]
    pool = None  # mp.Pool(processes=5)
    diffev_fit.fit(computational_pool=pool)
    no_plx_t0 = diffev_fit.fit_results['best_model'][0]

    pspl_plx = PSPL_model.PSPLmodel(your_event, parallax=['Full', no_plx_t0])

    diffev_fit = DE_fit.DEfit(pspl_plx, loss_function='chi2')
    diffev_fit.fit_parameters['u0'][1] = [0., 5.]
    diffev_fit.fit(computational_pool=pool)
    # Do gradient fit to fine tune solution
    gradient_fit = TRF_fit.TRFfit(pspl_plx, loss_function='chi2')
    gradient_fit.fit_parameters['u0'][1] = [0, 5.]  # PyLima has a short limit for u0

    gradient_fit.model_parameters_guess = diffev_fit.fit_results['best_model'][0:5]
    gradient_fit.fit()

    # results = []
    # chi2 = []

    results = gradient_fit.fit_results['best_model']
    chi2 = gradient_fit.fit_results['chi2']

    # Now fit - u0 and add to the results array
    # diffev_fit.fit_parameters['u0'][1] = [-5., 0.]
    # diffev_fit.fit(computational_pool=pool)
    # Do gradient fit to fine tune solution
    # gradient_fit = TRF_fit.TRFfit(pspl_plx, loss_function='chi2')
    # gradient_fit.fit_parameters['u0'][1] = [-5., 0.]  # PyLima has a short limit for u0

    # gradient_fit.model_parameters_guess = diffev_fit.fit_results['best_model'][0:5]
    # gradient_fit.fit()

    # results.append(gradient_fit.fit_results['best_model'])
    # chi2.append(gradient_fit.fit_results['chi2'])
    return results, chi2


def filter_by_q(event_path, pspl_pars, pspl_chi2, tstar, grid_q, grid_s, grid_alpha, a1_list, parallax=False,
                pspl_thresh=-50):
    """
    Finds best initial conditions on grid for each value of q.
    Does a very simple check for s>1 s<1, but this often will not find an s/1/s degeneracy.
    """
    nq = grid_q.shape[0]
    ns = grid_s.shape[0]
    na = grid_alpha.shape[0]
    n_models = nq * ns * na
    grid_file = np.loadtxt(f'{event_path}/Data/grid_fit.txt')
    if parallax:
        best_grid_models = np.zeros(shape=(nq, 11))
        # print(f'PSPL + PLX pars: {pspl_pars[0]} {pspl_pars[1]} {pspl_pars[2]} {pspl_pars[3]} {pspl_pars[4]}')
    else:
        best_grid_models = np.zeros(shape=(nq, 9))
        # print(f'PSPL pars: {pspl_pars[0]} {pspl_pars[1]} {pspl_pars[2]}')
    index = np.arange(0, n_models)
    chi2 = grid_file[:, 1]
    q_index, s_index, alpha_index = RECOVER_INDICES(index, nq, ns, na)

    print(q_index.dtype, s_index.dtype, alpha_index.dtype)
    q_values = grid_q[q_index]
    s_values = grid_s[s_index]
    alpha_values = grid_alpha[alpha_index]
    print('Picking best models for each mass ratio')
    q_unique = list(np.unique(q_values))
    # print(q_unique)
    best_grid_model_indices = []
    for q in q_unique:
        # print(q)
        subgrid = index[np.where(q_values == q)[0]]
        print(subgrid.shape)
        # filter results for 1 q value.
        best_model_argsort_index = np.argsort(chi2[subgrid])[0]  # get index of best chi2
        best_model_index = subgrid[best_model_argsort_index]  # get
        best_grid_model_indices.append(best_model_index)
    # pars = [np.log(fsblpars[0]), np.log(fsblpars[1]), psplpars[0], fsblpars[2],
    # np.log(fsblpars[3] / psplpars[1]), np.log(psplpars[1]), psplpars[2]]

    # now reconstruct actual inital condition
    for i in range(len(best_grid_model_indices)):
        ind = best_grid_model_indices[i]
        best_grid_models[i, 0] = s_values[ind]
        best_grid_models[i, 1] = q_values[ind]
        best_grid_models[i, 2] = pspl_pars[0]
        best_grid_models[i, 3] = alpha_values[ind]
        best_grid_models[i, 4] = tstar / pspl_pars[1]
        best_grid_models[i, 5] = pspl_pars[1]
        best_grid_models[i, 6] = pspl_pars[2]
        if parallax:
            best_grid_models[i, 7] = pspl_pars[3]
            best_grid_models[i, 8] = pspl_pars[4]
            best_grid_models[i, 9] = chi2[ind]
            best_grid_models[i, 10] = pspl_chi2 - chi2[ind]
        else:
            best_grid_models[i, 7] = chi2[ind]
            best_grid_models[i, 8] = pspl_chi2 - chi2[ind]

    print('Done')
    return best_grid_models


def filter_by_q_and_s(event_path, pspl_pars, pspl_chi2, tstar, grid_q, grid_s, grid_alpha,
                      parallax=False, pspl_thresh=0,
                      number_of_best_models_before_q_filter=13, number_of_final_models=20,
                      number_of_minimum_unique_q=4):
    """
    Finds best initial conditions on grid, with only one model allowed for each q and s.
    Tries to ensure 4 unique mass ratios, but will stop at 20 total initial conditions.
    Mini steps:
    1. Saves 13 best models with unique combination of q and s.
    2a. If < 4 unique mass ratios, check for 3 best models with different mass ratios, then 2, then 2 again.
    In a way there are 4 unique mass ratios.
    2b. If >= 4 unique mass ratios, add best models with unique q,s until 20 initial conditions.
    """

    # Number of models in the grid
    nq, ns, na = grid_q.shape[0], grid_s.shape[0], grid_alpha.shape[0]
    n_models = nq * ns * na
    # Read in grid file
    grid_file = np.loadtxt(f'{event_path}/Data/grid_fit.txt')
    chi2 = grid_file[:, 1]
    delta_chi2 = pspl_chi2 - chi2
    # Recover q,s,alpha values for each model
    index = np.arange(0, n_models)
    q_index, s_index, alpha_index = RECOVER_INDICES(index, nq, ns, na)
    print(q_index.dtype, s_index.dtype, alpha_index.dtype)
    q_values, s_values, alpha_values = grid_q[q_index], grid_s[s_index], grid_alpha[alpha_index]



    # Sort all models by chi2 (ascending)
    argsort_chi2_indices = np.argsort(chi2)
    delta_chi2_le_0_flag = 0  # flag to set if all models that improve the chi2 have been found.
    best_grid_model_indices = []
    best_grid_model_q_s_tuples = []

    # ----------------------
    # Step 1: select top 13 (number_of_best_models_before_q_filter) unique (q,s)
    # ----------------------
    print('Picking best models. 1 alpha per q,s pair allowed.')
    saved_skipped_indices = []
    last_step1_model_index = 0
    for model_index in range(n_models):
        if len(best_grid_model_indices) == number_of_best_models_before_q_filter:
            last_step1_model_index = model_index - 1  # Just to be true to the variable name
            #print(f'Step 1: {len(best_grid_model_indices)} initial conditions selected')
            break
        current_index = argsort_chi2_indices[model_index]
        q, s = q_values[current_index], s_values[current_index]
        if (q, s) in best_grid_model_q_s_tuples:
            saved_skipped_indices.append(model_index)
            continue
        else:
            if delta_chi2[current_index] > pspl_thresh:
                best_grid_model_indices.append(current_index)
                best_grid_model_q_s_tuples.append((q, s))
            else:
                print("No more models improve the chi2 over the pspl!")
                delta_chi2_le_0_flag = 1
                break
    # If we already have all models that improve chi2, print and go to best_model_indices part:
    if delta_chi2_le_0_flag:
        print(f'{len(best_grid_model_indices)} initial conditions selected')

    # ----------------------
    # Step 2: fill remaining slots to reach number_of_final_models (20)
    # ----------------------
    else:
        # Track unique q (mass ratios)
        q_in_init_conds = []
        for qs in best_grid_model_q_s_tuples:
            q_in_init_conds.append(qs[0])
        unique_q_in_init_conds = np.unique(q_in_init_conds)

        # if there are already 4 unique mass ratios, continue adding best (q,s) models
        if len(unique_q_in_init_conds) >= number_of_minimum_unique_q:
            start_i = last_step1_model_index + 1
            for remaining_model_index in range(start_i, n_models):
                if len(best_grid_model_indices) == number_of_final_models:
                    #print(f'Step 2: {len(best_grid_model_indices)} initial conditions selected')
                    break
                current_index = argsort_chi2_indices[remaining_model_index]
                s = s_values[current_index]
                q = q_values[current_index]
                if (q, s) in best_grid_model_q_s_tuples:
                    saved_skipped_indices.append(remaining_model_index)
                    continue
                else:
                    if delta_chi2[current_index] > pspl_thresh:
                        best_grid_model_indices.append(current_index)
                        best_grid_model_q_s_tuples.append((q, s))
                    else:
                        print("No more models improve the chi2 over the pspl!")
                        delta_chi2_le_0_flag = 1
                        break
        # however, if there are less than 4 unique mass ratios, try to add more
        else:
            print("Warning! Less than 4 unique mass ratios in the initial conditions")
            print("Adding more initial conditions")
            # batch_sizes = [3, 2, 2]
            # That's what we have /\ , but writing it more generic:
            number_of_models_that_are_needed = number_of_final_models - number_of_best_models_before_q_filter
            batch_sizes = [number_of_models_that_are_needed // (number_of_minimum_unique_q - 1) +
                           (i < (number_of_models_that_are_needed) % (number_of_minimum_unique_q - 1))
                           for i in range(number_of_minimum_unique_q - 1)]
            start_i = last_step1_model_index + 1
            nstop = number_of_best_models_before_q_filter
            for batch_counter in range(3):
                nstop += batch_sizes[batch_counter]
                #print('nstop',nstop)
                if len(unique_q_in_init_conds) < number_of_minimum_unique_q:
                    for new_model_index in range(start_i, n_models):
                        if len(best_grid_model_indices) == nstop:
                            # If batch is complete, it recalculates again the number of unique mass ratios
                            # and moves the start index to the last model added and breaks to start a new batch
                            q_in_init_conds = []
                            for qs in best_grid_model_q_s_tuples:
                                q_in_init_conds.append(qs[0])
                            unique_q_in_init_conds = np.unique(q_in_init_conds)
                            start_i = new_model_index
                            break
                        else:
                            # While batch is not complete, keep adding models
                            # if unique q is still not reached, only add models with new mass ratios:
                            current_index = argsort_chi2_indices[new_model_index]
                            q, s = q_values[current_index], s_values[current_index]
                            if ((q, s) in best_grid_model_q_s_tuples) or (q in unique_q_in_init_conds):
                                saved_skipped_indices.append(new_model_index)
                                continue
                            else:
                                if delta_chi2[current_index] > pspl_thresh:
                                    best_grid_model_indices.append(current_index)
                                    best_grid_model_q_s_tuples.append((q, s))
                                    #print(f'Step 3a: {len(best_grid_model_indices)} initial conditions selected')
                                else:
                                    print("No more models improve the chi2 over the pspl!")
                                    #print(f'Step 3b: {len(best_grid_model_indices)} initial conditions selected')
                                    delta_chi2_le_0_flag = 1
                                    break
                else:
                    print('Enough unique q added, adding skipped models.')
                    # if minimum unique q is reached, complete with next models from skipped models
                    # calculate how many more models are needed to reach 20
                    number_of_models_needed = number_of_final_models - len(best_grid_model_indices)
                    for skipped_index in range(0, number_of_models_needed):
                        if len(best_grid_model_indices) == nstop:
                            # If batch is complete, it breaks to start a new batch
                            #print(f'Step 3c: {len(best_grid_model_indices)} initial conditions selected')
                            break
                        else:
                            if skipped_index < len(saved_skipped_indices):
                                saved_index = saved_skipped_indices[skipped_index]
                            else:
                                saved_index = start_i  # fallback if no more skipped indices

                            current_index = argsort_chi2_indices[saved_index]
                            q, s = q_values[current_index], s_values[current_index]
                            if delta_chi2[current_index] > pspl_thresh:
                                best_grid_model_indices.append(current_index)
                                best_grid_model_q_s_tuples.append((q, s))
                                #print(f'Step 3d: {len(best_grid_model_indices)} initial conditions selected')
                            else:
                                print("No more models improve the chi2 over the pspl!")
                                #print(f'Step 3b: {len(best_grid_model_indices)} initial conditions selected')
                                delta_chi2_le_0_flag = 1
                                break

    # ----------------------
    # Step 3: populate best_grid_models array
    # ----------------------
    # now reconstruct actual initial conditions
    # Initialize best grid models array, with size depending on if static
    if parallax:
        best_grid_models = np.zeros(shape=(len(best_grid_model_indices), 11))
        # print(f'PSPL + PLX pars: {pspl_pars[0]} {pspl_pars[1]} {pspl_pars[2]} {pspl_pars[3]} {pspl_pars[4]}')
    else:
        best_grid_models = np.zeros(shape=(len(best_grid_model_indices), 9))
        # print(f'PSPL pars: {pspl_pars[0]} {pspl_pars[1]} {pspl_pars[2]}')

    for best_model_index in range(len(best_grid_model_indices)):
        ind = best_grid_model_indices[best_model_index]
        best_grid_models[best_model_index, 0] = s_values[ind]
        best_grid_models[best_model_index, 1] = q_values[ind]
        best_grid_models[best_model_index, 2] = pspl_pars[0]
        best_grid_models[best_model_index, 3] = alpha_values[ind]
        best_grid_models[best_model_index, 4] = tstar / pspl_pars[1]
        best_grid_models[best_model_index, 5] = pspl_pars[1]
        best_grid_models[best_model_index, 6] = pspl_pars[2]
        if parallax:
            best_grid_models[best_model_index, 7] = pspl_pars[3]
            best_grid_models[best_model_index, 8] = pspl_pars[4]
            best_grid_models[best_model_index, 9] = chi2[ind]
            best_grid_models[best_model_index, 10] = pspl_chi2 - chi2[ind]
        else:
            best_grid_models[best_model_index, 7] = chi2[ind]
            best_grid_models[best_model_index, 8] = pspl_chi2 - chi2[ind]

    print('Done')
    return best_grid_models


def combine_grid_files(event_path, parallel):
    data_path = f'{event_path}/Data'
    if parallel:
        if os.path.exists(f'{data_path}/grid_fit.txt'):
            os.remove(f'{data_path}/grid_fit.txt')
        grid_files = glob.glob(f'{data_path}/grid_fit*')
        dataframes = []
        for file in grid_files:
            print(file)
            #print(len(dataframes))
            dataframes.append(pd.read_csv(file, names=['index', 'chi2', 'cal_time'], sep='\s+'))
            os.remove(file) # remove now unecessary file
            #print(len(dataframes))
        #print(dataframes)
        grid_fit = pd.concat(dataframes)
        del dataframes
        grid_fit = grid_fit.sort_values('index')
        grid_fit.to_csv(f'{data_path}/grid_fit.txt', index=False, header=False, sep=' ')
    else:
        if os.path.exists(f'{data_path}/grid_fit0.txt'):
            shutil.move(f'{event_path}/Data/grid_fit0.txt', f'{event_path}/Data/grid_fit.txt')
        else:
            raise FileNotFoundError
    return None

def write_inital_conditions(event_path,initial_conditions,parallax=False):
    ninit_conds = initial_conditions.shape[0]
    if parallax:
        model = 'LX'
    else:
        model = 'LS'
    temppath = f'{event_path}/InitCond/InitCond{model}.tmp'
    initpath = f'{event_path}/InitCond/InitCond{model}.txt'
    with open(temppath,'w') as tempinit:
        with open(initpath, 'r') as oldinit:
            npeaks_nconds = oldinit.readline()
            npeaks = int(npeaks_nconds.split(' ')[0])
            npeaks_nconds = f'{npeaks} {ninit_conds}\n'
            tempinit.write(npeaks_nconds)
            # write peaks from old file
            for n in range(npeaks):
                peak = oldinit.readline()
                tempinit.write(peak)
    #now write initial conditons to file
    initial_conditions.iloc[:,0:7].to_csv(temppath,sep=' ', mode='a',header=False, index = False)
    os.remove(initpath)
    os.rename(temppath,initpath)
    return None



def run_event(event_path, dataset_list, grid_s, grid_q, grid_alpha, tstar, a1_list, pspl_thresh, processors,
              satellitedir, parallax=False, method='lm', use_saved_pspl=False):
    """ Wrapper Function to go from pspl_fit to final RTModel runs."""
    # First do the PSPL fit
    # results=pspl_fit_pyLIMA(event_path=event_path,dataset_list=dataset_list)
    # pspl_pars = [results['u0'],results['tE'],results['t0']]
    # pspl_chi2 = results['chi2']
    if use_saved_pspl:
        pspl = np.loadtxt(f'{event_path}/Data/pspl_pars.txt', delimiter=',')
        if parallax:
            pspl_pars = pspl[0:5]
        else:
            pspl_pars = pspl[0:3]
        pspl_chi2 = pspl[-1]
    else:
        if parallax:
            pspl_results, pspl_chi2 = psplPLX_fit_pyLIMA(event_path=event_path, dataset_list=dataset_list,
                                                         satellitedir=satellitedir)
            pspl_pars = [pspl_results[1], pspl_results[2], pspl_results[0] - 2_450_000, pspl_results[3],
                         pspl_results[4]]
        else:
            pspl_results, pspl_chi2 = pspl_fit_pyLIMA(event_path=event_path, dataset_list=dataset_list)
            pspl_pars = [pspl_results[1], pspl_results[2], pspl_results[0] - 2_450_000]
    # if method == 'lm':
    #    pspl_chi2 = pspl_results.cost*2
    # else: pspl_chi2 = pspl_results.chi2

    print(pspl_chi2)

    print(pspl_pars)
    # save pspl fit to a txt file
    #
    time0 = time.time()
    if processors == 1:
        parallel = False
        # grid_file_name = f'{event_path}/Data/grid_fit.txt'
        grid_fit(event_path=event_path, dataset_list=dataset_list, pspl_pars=pspl_pars,
                 grid_s=grid_s, grid_q=grid_q, grid_alpha=grid_alpha, tstar=tstar, a1_list=a1_list, pspl_chi2=pspl_chi2,
                 parallax=parallax)
    else:
        parallel = True
        grid_fit_parallelized(event_path=event_path, dataset_list=dataset_list, pspl_pars=pspl_pars,
                              grid_s=grid_s, grid_q=grid_q, grid_alpha=grid_alpha, tstar=tstar, a1_list=a1_list,
                              pspl_chi2=pspl_chi2,
                              parallax=parallax, use_croin=False, nprocessors=processors)
    time1 = time.time()
    print(f'ICGS time: {time1 - time0}')

    time0 = time.time()

    if parallax:
        names = ['log(s)', 'log(q)', 'u0', 'alpha', 'log(rho)', 'log(tE)', 't0', 'piEN', 'piEE', 'chi2sum',
                 'delta_pspl_chi2']
    else:
        names = ['log(s)', 'log(q)', 'u0', 'alpha', 'log(rho)', 'log(tE)', 't0', 'chi2sum', 'delta_pspl_chi2']

    # Now run these in RTModel
    # Have RTModel prints go to log not stdout
    rtm = RTModel.RTModel()
    rtm.set_processors(nprocessors=processors)
    rtm.set_event(event_path)
    # if os.path.exists(f'{event_path}/Nature.txt'):
    rtm.archive_run()
    # have to remove old stuff or it affects the InitConds for everything.
    archive_list = glob.glob(f'{event_path}/run-*')
    for archived_run in archive_list:
        shutil.rmtree(archived_run)
    # Write some outputs after clearing the directory
    with open(f'{event_path}/Data/pspl_pars.txt', 'w') as f:
        f.write(f'{pspl_pars[0]},{pspl_pars[1]},{pspl_pars[2]},{pspl_chi2}')
    # np.savetxt(fname=f'{event_path}/Data/ICGS_initconds.txt', X=init_conds)  # save init conds to a text file
    # np.savetxt(f'{event_path}/Data/grid_fit.txt', grid_fit_results) # change to npy or parquet later
    
    combine_grid_files(event_path, parallel=parallel)
    init_conds = filter_by_q_and_s(event_path, pspl_pars, pspl_chi2, tstar, grid_q, grid_s, grid_alpha, parallax,
                                   pspl_thresh=0)
    filtered_df = pd.DataFrame(init_conds, columns=names)
    np.savetxt(f'{event_path}/Data/ICGS_initconds_chi2.txt', filtered_df.values)

    if parallax:
        # nostatic = True
        modeltypes = ['PS', 'PX', 'LX', 'LO']
    else:
        # nostatic = False
        modeltypes = ['PS', 'PX', 'LS', 'LX', 'LO']

    rtm.set_satellite_dir(satellitedir=satellitedir)
    peak_threshold = 5
    rtm.set_processors(nprocessors=processors)
    rtm.set_event(event_path)
    rtm.set_satellite_dir(satellitedir=satellitedir)
    rtm.config_Reader(otherseasons=0, binning=1000000, renormalize=0)
    rtm.config_InitCond(usesatellite=1, peakthreshold=peak_threshold, modelcategories=modeltypes),  # nostatic=nostatic)
    rtm.Reader()
    rtm.InitCond()


    # Do FSPL fit for comparison

    if parallax:

        print('Launching PS Fits')
        rtm.launch_fits('PS')
        rtm.ModelSelector('PS')

        print('Launching PX Fits')
        rtm.launch_fits('PX')
        rtm.ModelSelector('PX')

        write_inital_conditions(event_path, filtered_df, parallax=parallax)
        print('Launching LX Fits')
        rtm.launch_fits('LX')
        rtm.ModelSelector('LX')


    else:
        print('Launching PS Fits')
        rtm.launch_fits('PS')
        rtm.ModelSelector('PS')

        print('Launching PX Fits')
        rtm.launch_fits('PX')
        rtm.ModelSelector('PX')

        write_inital_conditions(event_path, filtered_df, parallax=parallax)
        print('Launching LS Fits')
        rtm.launch_fits('LS')
        rtm.ModelSelector('LS')

        print('Launching LX fits')
        rtm.launch_fits('LX')
        rtm.ModelSelector('LX')

    print('Launching LO fits')
    rtm.launch_fits('LO')
    rtm.ModelSelector('LO')
    rtm.Finalizer()
    print('Done')
    time1 = time.time()
    print(f'RTModel time: {time1 - time0}')
    return None


def run_event_from_crash(event_path, processors, satellitedir):
    """ run from an error (user) to final RTModel runs."""

    # Now run these in RTModel
    # Have RTModel prints go to log not stdout
    time0 = time.time()
    init_conds = np.loadtxt(fname=f'{event_path}/Data/ICGS_initconds.txt')
    rtm = RTModel.RTModel()
    rtm.set_processors(nprocessors=processors)
    rtm.set_event(event_path)
    modeltypes = ['PS', 'LS', 'LX', 'LO']
    rtm.set_satellite_dir(satellitedir=satellitedir)
    peak_threshold = 5
    rtm.set_processors(nprocessors=processors)
    rtm.set_event(event_path)
    rtm.set_satellite_dir(satellitedir=satellitedir)
    rtm.config_Reader(otherseasons=0, binning=1000000, renormalize=0)
    rtm.config_InitCond(usesatellite=1, peakthreshold=peak_threshold, modelcategories=modeltypes)
    rtm.Reader()
    rtm.InitCond()
    # Do FSPL fit for comparison
    print('Launching PS Fits')
    rtm.launch_fits('PS')
    rtm.ModelSelector('PS')
    print('Launching LS Fits')
    num_init_cond = init_conds.shape[0]
    for n in range(num_init_cond):
        init_cond = list(init_conds[n, :])
        # launch each fit from the init conds
        rtm.LevMar(f'LSfit{n:03}', parameters=init_cond)
    rtm.ModelSelector('LS')
    print('Launching LX and LO fits')
    rtm.launch_fits('LX')
    rtm.ModelSelector('LX')
    rtm.launch_fits('LO')
    rtm.ModelSelector('LO')
    rtm.Finalizer()
    print('Done')
    time1 = time.time()
    print(f'RTModel time: {time1 - time0}')
    return None


def update_event(event_path, processors, satellitedir):
    """ update rtmodel run."""

    rtm = RTModel.RTModel()
    rtm.set_processors(nprocessors=processors)
    rtm.set_event(event_path)
    # if os.path.exists(f'{event_path}/Nature.txt'):
    rtm.archive_run()
    rtm.set_satellite_dir(satellitedir=satellitedir)
    peak_threshold = 5
    rtm.set_processors(nprocessors=processors)
    rtm.set_event(event_path)
    rtm.set_satellite_dir(satellitedir=satellitedir)
    rtm.config_Reader(otherseasons=0, binning=1000000, renormalize=0)
    rtm.config_InitCond(usesatellite=1, onlyupdate=True, oldmodels=10,
                        modelcategories=['PS', 'LX', 'LO']),  # nostatic=nostatic)
    rtm.run()
    return None


def calculate_magnifications_pspl(pars,data_list,data_mag,ndatasets,VBMInstance):
    """Calculate LC magnitudes in each passband provided for PSPL model"""
    magnitude_list = []
    source_flux_list = []
    blend_flux_list = []
    pars_log = [np.log(pars[0]), np.log(pars[1]), pars[2]]
    for i in range(ndatasets):
        data = data_list[i]
        magnifications, y1, y2 = VBMInstance.PSPLLightCurve(pars_log,data[:,-1])
        meas_flux = 10**(-0.4*data[:,0])
        meas_flux_err = 0.4*np.log(10)*10**(-0.4*data[:,0])*data[:,1]
        source_flux,blend_flux = minimize_linear_pars(meas_flux,meas_flux_err,magnifications)
        magnifications, y1, y2 = VBMInstance.PSPLLightCurve(pars_log,data_mag)
        sim_flux = np.array(magnifications)*source_flux+blend_flux
        source_flux_list.append(source_flux)
        blend_flux_list.append(blend_flux)
        sim_magnitudes = -2.5*np.log10(sim_flux)
        magnitude_list.append(sim_magnitudes)
    return magnitude_list,source_flux_list,blend_flux_list


def plot_pspl(pars,dataset,event_path):
    data = np.loadtxt(f'{event_path}/Data/{dataset}')
    VBMInstance = VBMicrolensing.VBMicrolensing()
    VBMInstance.RelTol = 1e-03
    VBMInstance.Tol = 1e-03
    tmag = np.linspace(data[0,-1],data[-1,-1],80000)
    magnitude_list, source_flux_list, blend_flux_list = calculate_magnifications_pspl(pars,[data],tmag,1,VBMInstance)
    magnitude = magnitude_list[0]
    fig, ax = plt.subplots(dpi=100, layout='tight')
    ax.plot(tmag, magnitude, zorder=10, label='PSPL Fit', color='black')
    #magnitude_list, source_flux_list, blend_flux_list = calculate_magnifications_pspl([0.6291636,50.20121621,pars[-1]], [data],tmag, 1, VBMInstance)
    #magnitude = magnitude_list[0]
    #ax.plot(tmag, magnitude, zorder=10, label='PSPL True', color='red')
    ax.errorbar(data[:, -1], y=data[:, 0], yerr=data[:, 1], marker='.', markersize=0.75,
                linestyle=' ', label='W146')
    ax.set_xlim(pars[-1]-1*pars[1],pars[-1]+1*pars[1])
    ax.yaxis.set_inverted(True)
    ax.legend()
    plt.savefig(f'{event_path}/pspl_plot.png')
    plt.show()
    return None



"""
This file uses JH's pyLIMA version from https://github.com/jonhulberg/pyLIMAbk , to save the MCMC as a hdf5 file
So we can relaunch in case of job time limit on the supercomputer
For testing, this current forces only on CPU
"""
import os
# restrict numpy multithreading
os.environ["OMP_NUM_THREADS"] = "1"
os.environ["OPENBLAS_NUM_THREADS"] = "1"
os.environ["MKL_NUM_THREADS"] = "1"
os.environ["VECLIB_MAXIMUM_THREADS"] = "1"
os.environ["NUMEXPR_NUM_THREADS"] = "1"
import csv
import time
import sys

import numpy as np
import pandas as pd

from pyLIMA import event
from pyLIMA import telescopes
from pyLIMA.models import FSBL_model
from pyLIMA.fits import MCMCHDF5_fit,TRF_fit
from pyLIMA.models import pyLIMA_fancy_parameters

from jasmine.classes_and_files_reader import RTModel_results_cls as rtm_results
from jasmine.files_organizer import ra_and_dec_conversions as radec
#from jasmine.files_organizer.new_data_files_for_pylima import rtmodel_data_to_pylima as new_data_files
#from jasmine.constants import limb_darkening_parameters

def main(general_path_for_rtmodel_run_,
         pylima_data_folder_,
         model_path_for_initial_conditions_,
         satellite_directory_,
         event_name_,
         number_of_processors_=5,
         number_of_steps_=5000,
         number_of_walkers_=2,
         run_name=''):
    # WHERE YOUR DATA IS LOCATED:
    data_folder_path = f'{general_path_for_rtmodel_run_}/Data'
    ### Create a new EVENT object and give it a name.
    # Create new files
    #new_data_files(data_folder_path=data_folder_path,
    #               data_output_folder_path=pylima_data_folder_,
    #               event_number=event_name_)

    # Load previous information from RTModel
    print('Loading previous information from RTModel...')
    # Using jasmine
    rtm_model = rtm_results.ModelResults(file_to_be_read=model_path_for_initial_conditions_)
    # Coordinates
    file_path = f'{data_folder_path}/event.coordinates'  # Replace with the actual file path
    event_coordinates = pd.read_csv(file_path, delim_whitespace=True, header=None, names=['RA_HMS', 'Dec_DMS'])

    # Convert RA and Dec to decimal degrees
    ra_deg = radec.ra_hms_to_deg(event_coordinates['RA_HMS'].iloc[0])
    dec_deg = radec.dec_dms_to_deg(event_coordinates['Dec_DMS'].iloc[0])

    your_event = event.Event(ra=ra_deg, dec=dec_deg)
    your_event.name = f'rep_sample_{event_name_}'

    ### Load up the data
    print('Loading data...')
    names = ['index','HJD','flux','err','sat']
    lc_to_fit = pd.read_csv(f'{general_path_for_rtmodel_run_}/LCToFit.txt', names=names,sep='\s+',skiprows=1)
    lc_to_fit['HJD']+=2_450_000 #pylima wants hjd not hjd - 2450000
    filters = []
    with open(f'{general_path_for_rtmodel_run_}/FilterToData.txt','r') as f:
        for line in f:
            s=line.strip('\n')
            filters.append(s)
    ald_coefficients = []
    with open(f'{general_path_for_rtmodel_run_}/Data/LimbDarkening.txt','r') as f:
        for line in f:
            s = float(line.strip('\n'))
            ald_coefficients.append(s)

    data_1 = lc_to_fit[lc_to_fit['sat']==1]
    data_2 = lc_to_fit[lc_to_fit['sat']==2]
    data_3 = lc_to_fit[lc_to_fit['sat']==3]
    index_1 = np.unique(data_1['index'])[0]
    index_2 = np.unique(data_2['index'])[0]
    index_3 = np.unique(data_3['index'])[0]
    ald_1 = ald_coefficients[index_1]
    ald_2 = ald_coefficients[index_2]
    ald_3 = ald_coefficients[index_3]
    print(index_3)



    ### IMPORTANT: Tell the code that Roman is in space!
    # read in ephemerides files
    names = ['time', 'RA','Dec', 'dist', 'deldot']
    t1_ephem = pd.read_csv(f'{satellite_directory_}/satellite1.txt',names=names,comment='$',sep = '\s+')
    t2_ephem = pd.read_csv(f'{satellite_directory_}/satellite2.txt', names=names, comment='$',sep = '\s+')
    t3_ephem = pd.read_csv(f'{satellite_directory_}/satellite3.txt', names=names, comment='$',sep = '\s+')

    data_2 = data_2.reset_index()
    data_3 = data_3.reset_index()
    data_1.loc[:,'HJD'] = t1_ephem.loc[:,'time']
    data_2.loc[:,'HJD'] = t2_ephem.loc[:,'time']
    data_3.loc[:,'HJD'] = t3_ephem.loc[:,'time']
    #print(data_3)
    #     [time,mag,err_mag]
    telescope_1 = telescopes.Telescope(name='RomanW146',
                                       camera_filter='W146',
                                       lightcurve=data_1[['HJD', 'flux', 'err']].values,
                                       lightcurve_names=['time', 'flux', 'err_flux'],
                                       lightcurve_units=['HJD', 'flux', 'flux'])
    telescope_2 = telescopes.Telescope(name='RomanZ087',
                                       camera_filter='Z087',
                                       lightcurve=data_2[['HJD', 'flux', 'err']].values,
                                       lightcurve_names=['time', 'flux', 'err_flux'],
                                       lightcurve_units=['HJD', 'flux', 'flux'])
    telescope_3 = telescopes.Telescope(name='RomanK213',
                                       camera_filter='K213',
                                       lightcurve=data_3[['HJD', 'flux', 'err']].values,
                                       lightcurve_names=['time', 'flux', 'err_flux'],
                                       lightcurve_units=['HJD', 'flux', 'flux'])




    #print(telescope_1.spacecraft_positions)

    telescope_1.location = 'Space'
    telescope_1.spacecraft_name = 'Roman'
    telescope_1.spacecraft_positions['photometry'] = t1_ephem[['time','RA','Dec','dist']].values
    telescope_2.location = 'Space'
    telescope_2.spacecraft_name = 'Roman'
    telescope_2.spacecraft_positions['photometry'] = t2_ephem[['time','RA', 'Dec', 'dist']].values
    telescope_3.location = 'Space'
    telescope_3.spacecraft_name = 'Roman'
    telescope_3.spacecraft_positions['photometry'] = t3_ephem[['time','RA', 'Dec', 'dist']].values
    #print(telescope_1.spacecraft_positions)
    # Adding limb darkening
    telescope_1.ld_a1 = ald_1
    telescope_2.ld_a1 = ald_2
    telescope_3.ld_a1 = ald_3


    ### Append these two telescope data sets to your EVENT object.
    print('Appending data...')
    your_event.telescopes.append(telescope_1)
    your_event.telescopes.append(telescope_2)
    your_event.telescopes.append(telescope_3)
    print(telescope_1.name,telescope_1.ld_a1)
    print(telescope_2.name, telescope_2.ld_a1)
    print(telescope_3.name, telescope_3.ld_a1)
    ### Define the survey telescope that you want to use to align all other data sets to.
    ### We recommend using the data set with the most measurements covering the gretest
    ### time span of observations:
    print('Defining survey telescope...')
    your_event.find_survey('RomanW146')

    ### Run a quick sanity check on your input.
    print('Checking event...')
    your_event.check_event()
    print(rtm_model.model_parameters.t0 + 2450000)
    ### CHOSE MODEL
    print('Choosing model...')
    # parallax should be close to t0
    #define log parameters with fancy parameters
    #my_pars = {'log_tE': 'tE','log_rho':'rho','log_mass_ratio':'mass_ratio','log_separation':'separation'}

    fancy = pyLIMA_fancy_parameters.StandardFancyParameters()
    if rtm_model.model_type == 'LS':
        model_to_mcmc = FSBL_model.FSBLmodel(your_event, parallax=['None', 0.0],
                                             orbital_motion=['None',0.0],fancy_parameters=fancy)
        guess_parameters = np.array(
            [rtm_model.model_parameters.t0 + 2450000, rtm_model.model_parameters.u0,
             np.log10(rtm_model.model_parameters.tE), np.log10(rtm_model.model_parameters.rho),
             np.log10(rtm_model.model_parameters.separation), np.log10(rtm_model.model_parameters.mass_ratio),
             rtm_model.model_parameters.alpha])


    elif  rtm_model.model_type == 'LX':
        model_to_mcmc = FSBL_model.FSBLmodel(your_event, parallax=['Full', rtm_model.model_parameters.t0 + 2450000],
                                             orbital_motion=['None', 0.0],fancy_parameters=fancy)
        guess_parameters = np.array(
            [rtm_model.model_parameters.t0 + 2450000, rtm_model.model_parameters.u0,
             np.log10(rtm_model.model_parameters.tE), np.log10(rtm_model.model_parameters.rho),
             np.log10(rtm_model.model_parameters.separation), np.log10(rtm_model.model_parameters.mass_ratio),
             rtm_model.model_parameters.alpha, rtm_model.model_parameters.piEN, rtm_model.model_parameters.piEE])

    elif rtm_model.model_type == 'LO':
        model_to_mcmc = FSBL_model.FSBLmodel(your_event, parallax=['Full', rtm_model.model_parameters.t0 + 2450000],
                                             orbital_motion=['Circular',rtm_model.model_parameters.t0 + 2450000],fancy_parameters=fancy)
        guess_parameters = np.array(
            [rtm_model.model_parameters.t0 + 2450000, rtm_model.model_parameters.u0,
             np.log10(rtm_model.model_parameters.tE), np.log10(rtm_model.model_parameters.rho),
             np.log10(rtm_model.model_parameters.separation), np.log10(rtm_model.model_parameters.mass_ratio),
             rtm_model.model_parameters.alpha, rtm_model.model_parameters.piEN, rtm_model.model_parameters.piEE,
             365.25*rtm_model.model_parameters.gamma1,365.25*rtm_model.model_parameters.gamma2,365.25*rtm_model.model_parameters.gammaz])
    else:
        print('Model not found - only supported LS, LO and LX ')
        return


    print('Gradient Fitting...')
    #my_fit = MCMC_fit.MCMCfit(model_to_mcmc, MCMC_links=number_of_steps_, MCMC_walkers=number_of_walkers_,
    #                          loss_function='chi2')
    gradient_fit = TRF_fit.TRFfit(model_to_mcmc,loss_function='chi2')
    print('Defining guess parameters...')
    gradient_fit.model_parameters_guess = guess_parameters
    print('Defining constraints...')
    #use very wide constraints.
    # my_fit.fit_parameters['t0'][1] = [2458000, 2464000]  # PyLima limits for t0 doesnt allow roman simulations
    gradient_fit.fit_parameters['u0'][1] = [-5.0, 5.0]  # PyLima has a short limit for u0
    gradient_fit.fit_parameters['log_rho'][1] = [-5.3, -0.3]  # PyLima has a low limit for rho
    gradient_fit.fit_parameters['log_separation'][1] = [-5.3, 2.]
    gradient_fit.fit_parameters['log_mass_ratio'][1] = [-7., 0.]
    print(rtm_model.model_type)
    if (rtm_model.model_type == 'LX') or (rtm_model.model_type == 'LO'):
        gradient_fit.fit_parameters['piEN'][1] = [-2.0, 2.0]
        gradient_fit.fit_parameters['piEE'][1] = [-2.0, 2.0]
        if rtm_model.model_type == 'LO':
            gradient_fit.fit_parameters['v_para'][1] = [-365.25, 365.25]
            gradient_fit.fit_parameters['v_perp'][1] = [-365.25, 365.25]
            gradient_fit.fit_parameters['v_radial'][1] = [1E-7, 365.25]

    #print(my_fit.model_chi2(parameters=guess_parameters))
    #print(gradient_fit.fit_parameters)

    gradient_fit.fit()#computational_pool=pool
    #my_fit.fit_outputs()
    print('TR Fit Done!')
    # File #1
    # Gather necessary data
    print('MCMC Fitting')
    hdf5path = f'{pylima_data_folder_}/{event_name_}_mcmc_chain{run_name}_{rtm_model.model_type}.hdf5'
    my_fit = MCMCHDF5_fit.MCMCfit(model_to_mcmc, MCMC_links=number_of_steps_, MCMC_walkers=number_of_walkers_,
                              loss_function='chi2',hdf5path=hdf5path)
    print('Defining guess parameters...')
    my_fit.model_parameters_guess = gradient_fit.fit_results['best_model']
    print('Defining constraints...')
    # use very wide constraints.
    # my_fit.fit_parameters['t0'][1] = [2458000, 2464000]  # PyLima limits for t0 doesnt allow roman simulations
    my_fit.fit_parameters['u0'][1] = [-5.0, 5.0]  # PyLima has a short limit for u0
    my_fit.fit_parameters['log_rho'][1] = [-5.3, -0.3]  # PyLima has a low limit for rho
    my_fit.fit_parameters['log_separation'][1] = [-5.3, 2.]
    my_fit.fit_parameters['log_mass_ratio'][1] = [-7., 0.]
    print(rtm_model.model_type)
    if (rtm_model.model_type == 'LX') or (rtm_model.model_type == 'LO'):
        my_fit.fit_parameters['piEN'][1] = [-2.0, 2.0]
        my_fit.fit_parameters['piEE'][1] = [-2.0, 2.0]
        if rtm_model.model_type == 'LO':
            my_fit.fit_parameters['v_para'][1] = [-365.25, 365.25]
            my_fit.fit_parameters['v_perp'][1] = [-365.25, 365.25]
            my_fit.fit_parameters['v_radial'][1] = [1E-7, 365.25]

    #pool = mul.Pool(processes=number_of_processors_)
    my_fit.fit(computational_pool=None)
    print('MCMC Done')
    event_name = my_fit.model.event.name
    ra = my_fit.model.event.ra
    dec = my_fit.model.event.dec
    east = my_fit.model.event.East
    north = my_fit.model.event.North
    parameters_names = list(my_fit.model.model_dictionnary.keys())
    best_model_ndarray = my_fit.fit_results['best_model']
    fit_time = my_fit.fit_results['fit_time']
    MCMC_links = my_fit.MCMC_links
    MCMC_walkers = my_fit.MCMC_walkers
    MCMC_chains_with_fluxes = my_fit.fit_results['MCMC_chains_with_fluxes']
    sampler = my_fit.fit_results['fit_object']
    #print(my_fit.model_chi2(parameters=best_model_ndarray))
    #print(my_fit.model_likelihood(parameters=best_model_ndarray))
    saving_results_path = pylima_data_folder_
    # Write File #1
    #pyLIMA_plots.plot_lightcurves(model_to_mcmc,guess_parameters)
    #my_fit.fit_outputs(bokeh_plot=True,bokeh_plot_name=f'{pylima_data_folder_}/{event_name}plot.html')
    print('Saving best model...')
    with open(f'{saving_results_path}/{event_name_}_best_model{run_name}_{rtm_model.model_type}.csv', 'w', newline='') as csvfile1:
        csvfile1.write(f"# Event name: {event_name}\n")
        csvfile1.write(f"# RA: {ra}\n")
        csvfile1.write(f"# Dec: {dec}\n")
        csvfile1.write(f"# East: {east}\n")
        csvfile1.write(f"# North: {north}\n")

        writer = csv.writer(csvfile1)
        # Write header
        writer.writerow(parameters_names)
        # Write data
        writer.writerow(best_model_ndarray)

    # Write File #2
    print('Saving MCMC chains...')
    with open(f'{saving_results_path}/{event_name_}_mcmc_chain{run_name}_{rtm_model.model_type}.csv', 'w', newline='') as csvfile2:
        csvfile2.write(f"# Event name: {event_name}\n")
        csvfile2.write(f"# RA: {ra}\n")
        csvfile2.write(f"# Dec: {dec}\n")
        csvfile2.write(f"# East: {east}\n")
        csvfile2.write(f"# North: {north}\n")
        csvfile2.write(f"# Fit time: {fit_time}\n")
        csvfile2.write(f"# MCMC links: {MCMC_links}\n")
        csvfile2.write(f"# MCMC walkers: {MCMC_walkers}\n")
        csvfile2.write(f'#Autocorr Time = {sampler.get_autocorr_time(tol=0)}')
        csvfile2.write(f'#Acceptance Ratio =  {sampler.acceptance_fraction}')
        writer = csv.writer(csvfile2)
        # Write header
        writer.writerow(parameters_names + ['likelihood','prior'])
        # Write data
        for quoted_arrays in MCMC_chains_with_fluxes:
            for array in quoted_arrays:
                writer.writerow(array)
    print(f'Autocorr Time = {sampler.get_autocorr_time(tol=0)}')
    print(f'Acceptance Ratio =  {sampler.acceptance_fraction}')
    print('All done!')


def get_solutions(general_path,event_name,model_type):
    best_of_each_binary = pd.read_csv(f'{general_path}/{event_name}/Models/chi2_top1_of_each_binary_lens_model.csv')
    soln_name = best_of_each_binary[best_of_each_binary['model'].str.contains(model_type)]
    model_name_ = soln_name.iloc[0,0]
    print('Model name:', model_name_)
    return model_name_


if __name__ == '__main__':
    start = time.time()


    #     gs66-ponyta:
    # general_path_for_rtmodel_run = (f'/Users/sishitan/Documents/Scripts/RTModel_project/'
    #                                 f'rtmodel_pylima/datachallenge_events/event_{event_number:03}/')
    # pylima_data_folder = ('/Users/sishitan/Documents/Scripts/RTModel_project/'
    #                       'rtmodel_pylima/datachallenge_events/pylima_format_data')
    # pylima_data_folder = '/local/data/emussd1/greg_shared/pylima_mcmc/outputsx'
    # satellite_directory = '/local/data/emussd1/greg_shared/pylima_mcmc/satellitedir'
    # ############################

    # ###########################
    # discover
    general_path = '/discover/nobackup/sishitan/orbital_task/RTModel_runs/sample_rtmodel_v2.4'
    pylima_data_folder = '/discover/nobackup/sishitan/orbital_task/pylima_mcmcs'
    satellite_directory = '/discover/nobackup/sishitan/orbital_task/data/satellitedir'

    event_subrun = int(sys.argv[1])
    event_field = int(sys.argv[2])
    event_id = int(sys.argv[3])
    event_name = f'event_{event_subrun}_{event_field}_{event_id}'
    print('Event name:', event_name)
    model_type = sys.argv[4]
    model_name = get_solutions(general_path, event_name, model_type)
    number_of_process = 1
    number_of_steps = 5000
    number_of_walkers = 2

    # where the data in pylima format will be saved
    general_path_for_rtmodel_run = f'{general_path}/{event_name}'
    model_path_for_initial_conditions = f'{general_path_for_rtmodel_run}/Models/{model_name}.txt'
    main(event_name_=event_name,
         general_path_for_rtmodel_run_=general_path_for_rtmodel_run,
         pylima_data_folder_=pylima_data_folder,
         model_path_for_initial_conditions_=model_path_for_initial_conditions,
         satellite_directory_=satellite_directory,
         number_of_processors_=number_of_process,
         number_of_steps_=number_of_steps,
         number_of_walkers_=number_of_walkers,
         run_name=f'_{number_of_steps}_{number_of_walkers}_{model_name}')
    finish = time.time()
    print('Pylima MCMC took', finish - start, f'seconds for event {event_name} using {number_of_process} processors,'
                                              f'{number_of_steps} number_of_steps '
                                              f'and {number_of_walkers} number_of_walkers.')



import pandas as pd
from jasmine import ModelResults


def get_LO_names(*, summary_path_folder_, run_type_, event_unique_id_):
    summary_df = pd.read_csv(f'{summary_path_folder_}/{run_type_}_LX_LO_summary.csv')
    LO_name = summary_df[summary_df['event_name'] == f'event_{event_unique_id_}']['best_LO'].values[0]
    return LO_name


def LO_model_retriever_per_event_and_runtype(*, general_path_, run_type_, event_unique_id_):
    summary_path_folder = f'{general_path_}/RTModel_runs'
    LO_name = get_LO_names(summary_path_folder_=summary_path_folder,
                           run_type_=run_type_,
                           event_unique_id_=event_unique_id_)
    LO_model = ModelResults(f'{summary_path_folder}/{run_type_}/event_{event_unique_id_}/Models/{LO_name}.txt')
    # Convert model parameters (dataclass) to dictionary
    lo_data = vars(LO_model.model_parameters)
    # Convert to DataFrame
    lo_df = pd.DataFrame(lo_data, index=pd.Index([f'event_{event_unique_id_}'], name='event_name'))
    return lo_df


def LO_model_orbital_motion_parameters_to_csv(*, run_type_, general_path_):
    all_dataframes = []

    all_representative_sample_df = pd.read_csv(f'{general_path_}/data/gulls_orbital_motion_extracted/'
                                               f'OMPLDG_croin_cassan.sample.csv')
    all_representative_sample_df['event_unique_id'] = (all_representative_sample_df['SubRun'].astype(str) + "_" +
                                                       all_representative_sample_df['Field'].astype(str) + "_" +
                                                       all_representative_sample_df['EventID'].astype(str))

    for event_unique_id_ in all_representative_sample_df['event_unique_id']:
        try:
            lo_df = LO_model_retriever_per_event_and_runtype(general_path_=general_path_,
                                                             run_type_=run_type_,
                                                             event_unique_id_=event_unique_id_)
            all_dataframes.append(lo_df)
        except Exception as e:
            print(f"Skipping run_type '{run_type_}' for event '{event_unique_id_}': {e}")

    # Save to CSV
    combined_df = pd.concat(all_dataframes)
    lo_csv_path = f'{general_path_}/RTModel_runs/{run_type_}_LO_model_parameters.csv'
    combined_df.to_csv(lo_csv_path)



if __name__ == '__main__':
    run_type = 'sample_rtmodel_v2.4'
    # general_path = '/Users/stela/Documents/Scripts/orbital_task'
    general_path = '/discover/nobackup/sishitan/orbital_task'

    LO_model_orbital_motion_parameters_to_csv(general_path_=general_path,
                                              run_type_=run_type)




import glob

import VBMicrolensing
import numpy as np
import matplotlib.pyplot as plt
from gridfit import *
#[np.log(fsblpars[0]), np.log(fsblpars[1]), psplpars[0], fsblpars[2],
#                np.log(fsblpars[3] / psplpars[1]), np.log(psplpars[1]), psplpars[2]]
def make_comparison_plot(model_type,event_path,initconds_type,xrange,yrange):
    a1_list = [0.33,0.4534,0.2517] # limb darkening
    # load data
    dataset_list = ['RomanW146sat1.dat','RomanZ087sat2.dat','RomanK213sat3.dat']
    ndatasets = len(dataset_list)
    data_list = []
    for i in range(ndatasets):
        data_list.append(np.loadtxt(f'{event_path}/Data/{dataset_list[i]}'))
    ###

    event_name = event_path.split('/')[-1]
    VBMInstance = VBMicrolensing.VBMicrolensing()
    if initconds_type == 'icgs':
        icgs_init_conds = np.loadtxt(f'{event_path}/Data/ICGS_initconds_chi2.txt')#_chi2
        #print(icgs_init_conds)
        model_files_list = glob.glob(f'{event_path}/Models/*{model_type}*')
        models_list = []
        initcond_indices = []
        for file in model_files_list:
            file = file.split('/')[-1]
            models_list.append(file)
            index = int(file[3:6])
            initcond_indices.append(index)
        for i in range(len(models_list)):
            model_path = model_files_list[i]
            model_name = models_list[i]
            icgs_index = initcond_indices[i]
            init_cond = icgs_init_conds[icgs_index,:]
            pars = init_cond[0:7]

            log_pars = np.copy(pars)
            log_pars[0] = np.log(pars[0])
            log_pars[1] = np.log(pars[1])
            log_pars[4] = np.log(pars[4])
            log_pars[5] = np.log(pars[5])


            tmag = np.linspace(pars[-1] - 5 * pars[5], pars[-1] + 5 * pars[5], 80000)

            magnitude_list_init, flux_list_init, fluxdata_list_init, fluxerr_list_init, source_flux_list_init, blend_flux_list_init = calculate_magnifications_plots(log_pars, a1_list, data_list, tmag, ndatasets, VBMInstance, parallax=False)
            fig, ax = plt.subplots(dpi=100, layout='tight')

            ax.errorbar(data_list[0][:, -1], y=data_list[0][:, 0], yerr=data_list[0][:, 1], marker='.', markersize=0.75,
                        linestyle=' ', label='W146',color='springgreen')

            ax.plot(tmag, magnitude_list_init[0], zorder=10, linestyle ='--', label='Initial Condition', color='red')


            if xrange is not None:
                ax.set_xlim(xrange[0],xrange[1])
                ax.set_ylim(yrange[0],yrange[1])
            else:ax.set_xlim(tmag[0],tmag[-1])

            ax.yaxis.set_inverted(True)
            ax.set_xlabel('HJD - 2450000')
            ax.set_ylabel('W146')

            # read in RTModel solution
            rtm_pars = np.loadtxt(model_path, max_rows=1)
            log_rtm_pars = np.copy(rtm_pars)
            log_rtm_pars[0] = np.log(rtm_pars[0])
            log_rtm_pars[1] = np.log(rtm_pars[1])
            log_rtm_pars[4] = np.log(rtm_pars[4])
            log_rtm_pars[5] = np.log(rtm_pars[5])
            magnitude_list_rtm, flux_list_rtm, fluxdata_list_rtm, fluxerr_list_rtm, source_flux_list_rtm, blend_flux_list_rtm = calculate_magnifications_plots(log_rtm_pars[0:7], a1_list, data_list, tmag, ndatasets,
                                                                  VBMInstance, parallax=False)
            ax.plot(tmag, magnitude_list_rtm[0], zorder=10, linestyle='-', label='Levenberg-Marquardt', color='black')
            ax.legend()
            title = rf'{event_name} - {model_name}\
                     \nInitial Condition: s = {pars[0]:.3f}, q = {pars[0]:.3f}, $\chi^2$ = {init_cond[-1]:.3f}\
            \nFit: s = {rtm_pars[0]:.3f}, q = {rtm_pars[0]:.3f}, $\chi^2$ = {rtm_pars[-1]:.3f}'

            ax.set_title(f'{event_name} - {model_name}')
            plt.show()
    else:
        print('not implemented')
        return None # for now

if __name__ == "__main__":
    model_type_ = 'LS'
    event_path_='/Users/jmbrashe/VBBOrbital/ICGS_analysis/benchmark_parallel/event_0_6_24'
    initconds_type = 'icgs'
    make_comparison_plot(model_type=model_type_,event_path=event_path_,initconds_type=initconds_type,xrange=None,yrange=None)